from common.constants_ac import *

# some constants from MFSCommunication.h
ANTOAN_NOP = 0

ANTOCS_CLEAR_ERRORS        = (PROTO_BASE+306)
CSTOAN_CLEAR_ERRORS        = (PROTO_BASE+307)

CLTOMA_CSERV_LIST          = (PROTO_BASE+500)
MATOCL_CSERV_LIST          = (PROTO_BASE+501)
CLTOAN_CHART_DATA          = (PROTO_BASE+506)
ANTOCL_CHART_DATA          = (PROTO_BASE+507)
CLTOMA_SESSION_LIST        = (PROTO_BASE+508)
MATOCL_SESSION_LIST        = (PROTO_BASE+509)
CLTOMA_INFO                = (PROTO_BASE+510)
MATOCL_INFO                = (PROTO_BASE+511)
CLTOMA_FSTEST_INFO         = (PROTO_BASE+512)
MATOCL_FSTEST_INFO         = (PROTO_BASE+513)
CLTOMA_CHUNKSTEST_INFO     = (PROTO_BASE+514)
MATOCL_CHUNKSTEST_INFO     = (PROTO_BASE+515)
CLTOMA_CHUNKS_MATRIX       = (PROTO_BASE+516)
MATOCL_CHUNKS_MATRIX       = (PROTO_BASE+517)
CLTOMA_QUOTA_INFO          = (PROTO_BASE+518)
MATOCL_QUOTA_INFO          = (PROTO_BASE+519)
CLTOMA_EXPORTS_INFO        = (PROTO_BASE+520)
MATOCL_EXPORTS_INFO        = (PROTO_BASE+521)
CLTOMA_MLOG_LIST           = (PROTO_BASE+522)
MATOCL_MLOG_LIST           = (PROTO_BASE+523)
CLTOMA_CSSERV_COMMAND      = (PROTO_BASE+524)
MATOCL_CSSERV_COMMAND      = (PROTO_BASE+525)
CLTOMA_SESSION_COMMAND     = (PROTO_BASE+526)
MATOCL_SESSION_COMMAND     = (PROTO_BASE+527)
CLTOMA_MEMORY_INFO         = (PROTO_BASE+528)
MATOCL_MEMORY_INFO         = (PROTO_BASE+529)
CLTOMA_LIST_OPEN_FILES     = (PROTO_BASE+532)
MATOCL_LIST_OPEN_FILES     = (PROTO_BASE+533)
CLTOMA_LIST_ACQUIRED_LOCKS = (PROTO_BASE+534)
MATOCL_LIST_ACQUIRED_LOCKS = (PROTO_BASE+535)
CLTOMA_MASS_RESOLVE_PATHS  = (PROTO_BASE+536)
MATOCL_MASS_RESOLVE_PATHS  = (PROTO_BASE+537)
CLTOMA_SCLASS_INFO         = (PROTO_BASE+542)
MATOCL_SCLASS_INFO         = (PROTO_BASE+543)
CLTOMA_MISSING_CHUNKS      = (PROTO_BASE+544)
MATOCL_MISSING_CHUNKS      = (PROTO_BASE+545)
CLTOMA_PATTERN_INFO        = (PROTO_BASE+548)
MATOCL_PATTERN_INFO        = (PROTO_BASE+549)
CLTOMA_INSTANCE_NAME       = (PROTO_BASE+550)
MATOCL_INSTANCE_NAME       = (PROTO_BASE+551)

CLTOCS_HDD_LIST            = (PROTO_BASE+600)
CSTOCL_HDD_LIST            = (PROTO_BASE+601)
MFS_MESSAGE = 1

FEATURE_EXPORT_UMASK        = 0
FEATURE_EXPORT_DISABLES     = 1
FEATURE_SESSION_STATS_28    = 2
FEATURE_INSTANCE_NAME       = 4
FEATURE_CSLIST_MODE         = 5
FEATURE_SCLASS_IN_MATRIX    = 7
FEATURE_DEFAULT_GRACEPERIOD = 8
FEATURE_LABELMODE_OVERRIDES = 9
FEATURE_SCLASSGROUPS        = 10

MASKORGROUP = 4
SCLASS_EXPR_MAX_SIZE = 128

MFS_CSSERV_COMMAND_REMOVE         = 0
MFS_CSSERV_COMMAND_BACKTOWORK     = 1
MFS_CSSERV_COMMAND_MAINTENANCEON  = 2
MFS_CSSERV_COMMAND_MAINTENANCEOFF = 3
MFS_CSSERV_COMMAND_TMPREMOVE      = 4

MFS_SESSION_COMMAND_REMOVE = 0

PATTERN_EUGID_ANY            = 0xFFFFFFFF
PATTERN_OMASK_SCLASS         = 0x01
PATTERN_OMASK_TRASHRETENTION = 0x02
PATTERN_OMASK_EATTR          = 0x04

SCLASS_ARCH_MODE_CTIME      = 0x01
SCLASS_ARCH_MODE_MTIME      = 0x02
SCLASS_ARCH_MODE_ATIME      = 0x04
SCLASS_ARCH_MODE_REVERSIBLE = 0x08
SCLASS_ARCH_MODE_FAST       = 0x10
SCLASS_ARCH_MODE_CHUNK      = 0x20

STATUS_OK      = 0
ERROR_NOTFOUND = 41
ERROR_ACTIVE   = 42

STATE_UNREACHABLE = 0xFFFF
STATE_DUMMY       = 0x00
STATE_LEADER      = 0x01
STATE_ELECT       = 0x02
STATE_FOLLOWER    = 0x03
STATE_USURPER     = 0x04
STATE_DEPUTY      = 0x05
STATE_MASTERCE    = 0xFF
STATE_STR_WRONGANSWER = "WRONG ANSWER"
STATE_STR_UNREACHABLE = "UNREACHABLE"
STATE_STR_BUSY        = "BUSY"

#CSTOCL_HDD_LIST flags constants
CS_HDD_MFR      = 0x01
CS_HDD_DAMAGED  = 0x02
CS_HDD_SCANNING = 0x04
CS_HDD_INVALID  = 0x08

MFRSTATUS_VALIDATING = 0 # unknown after disconnect or creation or unknown, loop in progress
MFRSTATUS_INPROGRESS = 1 # chunks still needs to be replicated, can't be removed 
MFRSTATUS_READY      = 2 # can be removed, whole loop has passed

CS_HDD_CS_VALID       = 0 #valid hdd data from CS
CS_HDD_CS_TOO_OLD     = 1 #too old CS to get info on its discs
CS_HDD_CS_UNREACHABLE = 2 #can't connect to CS and get info on its discs

CS_LOAD_NORMAL	       = 0
CS_LOAD_OVERLOADED     = 1
CS_LOAD_REBALANCE      = 2
CS_LOAD_FAST_REBALANCE = 3
CS_LOAD_GRACEFUL       = 4

LASTSTORE_UNKNOWN		 = 0 # if laststore_ts==0
LASTSTORE_META_STORED_BG = 0 # Stored in background if laststore_ts>0
LASTSTORE_DOWNLOADED     = 1 # Downloaded
LASTSTORE_META_STORED_FG = 2 # Stored in foreground
LASTSTORE_CRC_STORED_BG  = 3 # CRC stored in background

MISSING_CHUNK_TYPE_NOCOPY         = 0
MISSING_CHUNK_TYPE_INVALID_COPIES = 1
MISSING_CHUNK_TYPE_WRONG_VERSIONS = 2
MISSING_CHUNK_TYPE_PARTIAL_EC     = 3

FSTEST_INFO_STR_LIMIT = 100000 # 100k limit of datastr in MATOCL_FSTEST_INFO

UNRESOLVED = "(unresolved)"

THRESHOLD_WARNING = 0.90
THRESHOLD_ERROR   = 0.95

LICVER_CE = -1 # Community Edition licver value

AJAX_NONE      = None
AJAX_CONTAINER = 1
AJAX_METRICS   = 2

# Chunk matrix summary columns
MX_COL_MISSING       = 0
MX_COL_ENDANGERED    = 1
MX_COL_UNDERGOAL     = 2
MX_COL_STABLE        = 3
MX_COL_OVERGOAL      = 4
MX_COL_DELETEPENDING = 5
MX_COL_DELETEREADY   = 6
MX_COL_TOTAL         = 7

MFS_LOCK_TYPE_UNKNOWN   = 0
MFS_LOCK_TYPE_SHARED		= 1
MFS_LOCK_TYPE_EXCLUSIVE	= 2



from common.utils import *
from common.utilsgui import *
from common.models import *

def render(dp, fields, vld):
	QUorder = fields.getint("QUorder", 0)
	QUrev = fields.getint("QUrev", 0)

	out = []
	out.append("""<div class="tab_title">Active quotas</div>""")
	out.append("""<table class="acid_tab acid_tab_zebra_C1_C2 acid_tab_storageid_mfsquota">""")
	out.append("""	<tr>""")
	out.append("""		<th rowspan="3" class="acid_tab_enumerate">#</th>""")
	out.append("""		<th rowspan="3">Path</th>""")
	out.append("""	<th colspan="6">Soft&nbsp;quota</th>""")
	out.append("""	<th colspan="4">Hard&nbsp;quota</th>""")
	out.append("""	<th colspan="12">Current&nbsp;values</th>""")
	out.append("""	</tr>""")
	out.append("""	<tr>""")
	out.append("""		<th rowspan="2" class="wrap">grace period</th>""")
	out.append("""		<th rowspan="2" class="wrap">time to expire</th>""")
	out.append("""		<th rowspan="2">inodes</th>""")
	out.append("""		<th rowspan="2">length</th>""")
	out.append("""		<th rowspan="2">size</th>""")
	out.append("""		<th rowspan="2" class="wrap">real size</th>""")
	out.append("""		<th rowspan="2">inodes</th>""")
	out.append("""		<th rowspan="2">length</th>""")
	out.append("""		<th rowspan="2">size</th>""")
	out.append("""		<th rowspan="2" class="wrap">real size</th>""")
	out.append("""		<th colspan="3">inodes</th>""")
	out.append("""		<th colspan="3">length</th>""")
	out.append("""		<th colspan="3">size</th>""")
	out.append("""		<th colspan="3">real&nbsp;size</th>""")
	out.append("""	</tr>""")
	out.append("""	<tr>""")
	out.append("""		<th>value</th>""")
	out.append("""		<th>% soft</th>""")
	out.append("""		<th>% hard</th>""")
	out.append("""		<th>value</th>""")
	out.append("""		<th>% soft</th>""")
	out.append("""		<th>% hard</th>""")
	out.append("""		<th>value</th>""")
	out.append("""		<th>% soft</th>""")
	out.append("""		<th>% hard</th>""")
	out.append("""		<th>value</th>""")
	out.append("""		<th>% soft</th>""")
	out.append("""		<th>% hard</th>""")
	out.append("""	</tr>""")


	maxperc,quotas = dp.get_quotas(QUorder, QUrev)
	for qu in quotas:
		graceperiod_default = 0
		if dp.master().has_feature(FEATURE_DEFAULT_GRACEPERIOD):
			if qu.exceeded & 2:
				qu.exceeded &= 1
				graceperiod_default = 1

		out.append("""	<tr>""")
		out.append("""		<td align="right"></td>""")
		out.append("""		<td align="left">%s</td>""" % htmlentities(qu.path))
		if qu.graceperiod>0:
			out.append("""		<td align="center"><span class="sortkey">%u </span><a style="cursor:default" title="%s">%s%s</a></td>""" % (qu.graceperiod,timeduration_to_fullstr(qu.graceperiod),timeduration_to_shortstr(qu.graceperiod)," (default)" if graceperiod_default else ""))
		else:
			out.append("""		<td align="center"><span class="sortkey">0 </span>default</td>""")
		if qu.timetoblock<0xFFFFFFFF:
			if qu.timetoblock>0:
				out.append("""		<td align="center"><span class="SEXCEEDED"><span class="sortkey">%u </span><a style="cursor:default" title="%s">%s</a></span></td>""" % (qu.timetoblock,timeduration_to_fullstr(qu.timetoblock),timeduration_to_shortstr(qu.timetoblock)))
			else:
				out.append("""		<td align="center"><span class="EXCEEDED"><span class="sortkey">0 </span>expired</span></td>""")
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span>-</td>""")
		if qu.qflags&1:
			out.append("""		<td align="right"><span>%s</span></td>""" % decimal_number_html(qu.sinodes))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span>-</td>""")
		if qu.qflags&2:
			out.append("""		<td align="right"><span class="sortkey">%u </span><a style="cursor:default" title="%s B"><span>%s</span></a></td>""" % (qu.slength,decimal_number(qu.slength),humanize_number(qu.slength,"&nbsp;")))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span>-</td>""")
		if qu.qflags&4:
			out.append("""		<td align="right"><span class="sortkey">%u </span><a style="cursor:default" title="%s B"><span>%s</span></a></td>""" % (qu.ssize,decimal_number(qu.ssize),humanize_number(qu.ssize,"&nbsp;")))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span>-</td>""")
		if qu.qflags&8:
			out.append("""		<td align="right"><span class="sortkey">%u </span><a style="cursor:default" title="%s B"><span>%s</span></a></td>""" % (qu.srealsize,decimal_number(qu.srealsize),humanize_number(qu.srealsize,"&nbsp;")))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span>-</td>""")
		if qu.qflags&16:
			out.append("""		<td align="right"><span>%s</span></td>""" % decimal_number_html(qu.hinodes))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span>-</td>""")
		if qu.qflags&32:
			out.append("""		<td align="right"><span class="sortkey">%u </span><a style="cursor:default" title="%s B"><span>%s</span></a></td>""" % (qu.hlength,decimal_number(qu.hlength),humanize_number(qu.hlength,"&nbsp;")))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span>-</td>""")
		if qu.qflags&64:
			out.append("""		<td align="right"><span class="sortkey">%u </span><a style="cursor:default" title="%s B"><span>%s</span></a></td>""" % (qu.hsize,decimal_number(qu.hsize),humanize_number(qu.hsize,"&nbsp;")))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span>-</td>""")
		if qu.qflags&128:
			out.append("""		<td align="right"><span class="sortkey">%u </span><a style="cursor:default" title="%s B"><span>%s</span></a></td>""" % (qu.hrealsize,decimal_number(qu.hrealsize),humanize_number(qu.hrealsize,"&nbsp;")))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span>-</td>""")
		out.append("""		<td align="right">%s</td>""" % decimal_number_html(qu.cinodes))
		if qu.qflags&1:
			if qu.sinodes>0:
				if qu.sinodes>=qu.cinodes:
					clz="NOTEXCEEDED"
				elif qu.timetoblock>0:
					clz="SEXCEEDED"
				else:
					clz="EXCEEDED"
				out.append("""		<td align="right"><span class="%s">%.2f</span></td>""" % (clz,(100.0*qu.cinodes)/qu.sinodes))
			else:
				out.append("""		<td align="right"><span class="sortkey">%.2f </span><span class="EXCEEDED">inf</span></td>""" % (maxperc))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span>-</td>""")
		if qu.qflags&16:
			if qu.hinodes>0:
				if qu.hinodes>qu.cinodes:
					clz="NOTEXCEEDED"
				else:
					clz="EXCEEDED"
				out.append("""		<td align="right"><span class="%s">%.2f</span></td>""" % (clz,(100.0*qu.cinodes)/qu.hinodes))
			else:
				out.append("""		<td align="right"><span class="sortkey">%.2f </span><span class="EXCEEDED">inf</span></td>""" % (maxperc))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span>-</td>""")
		out.append("""		<td align="right"><span class="sortkey">%u </span><a style="cursor:default" title="%s B">%s</a></td>""" % (qu.clength,decimal_number(qu.clength),humanize_number(qu.clength,"&nbsp;")))
		if qu.qflags&2:
			if qu.slength>0:
				if qu.slength>=qu.clength:
					clz="NOTEXCEEDED"
				elif qu.timetoblock>0:
					clz="SEXCEEDED"
				else:
					clz="EXCEEDED"
				out.append("""		<td align="right"><span class="%s">%.2f</span></td>""" % (clz,(100.0*qu.clength)/qu.slength))
			else:
				out.append("""		<td align="right"><span class="sortkey">%.2f </span><span class="EXCEEDED">inf</span></td>""" % (maxperc))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span>-</td>""")
		if qu.qflags&32:
			if qu.hlength>0:
				if qu.hlength>qu.clength:
					clz="NOTEXCEEDED"
				else:
					clz="EXCEEDED"
				out.append("""		<td align="right"><span class="%s">%.2f</span></td>""" % (clz,(100.0*qu.clength)/qu.hlength))
			else:
				out.append("""		<td align="right"><span class="sortkey">%.2f </span><span class="EXCEEDED">inf</span></td>""" % (maxperc))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span>-</td>""")
		out.append("""		<td align="right"><span class="sortkey">%u </span><a style="cursor:default" title="%s B">%s</a></td>""" % (qu.csize,decimal_number(qu.csize),humanize_number(qu.csize,"&nbsp;")))
		if qu.qflags&4:
			if qu.ssize>0:
				if qu.ssize>=qu.csize:
					clz="NOTEXCEEDED"
				elif qu.timetoblock>0:
					clz="SEXCEEDED"
				else:
					clz="EXCEEDED"
				out.append("""		<td align="right"><span class="%s">%.2f</span></td>""" % (clz,(100.0*qu.csize)/qu.ssize))
			else:
				out.append("""		<td align="right"><span class="sortkey">%.2f </span><span class="EXCEEDED">inf</span></td>""" % (maxperc))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span>-</td>""")
		if qu.qflags&64:
			if qu.hsize>0:
				if qu.hsize>qu.csize:
					clz="NOTEXCEEDED"
				else:
					clz="EXCEEDED"
				out.append("""		<td align="right"><span class="%s">%.2f</span></td>""" % (clz,(100.0*qu.csize)/qu.hsize))
			else:
				out.append("""		<td align="right"><span class="sortkey">%.2f </span><span class="EXCEEDED">inf</span></td>""" % (maxperc))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span>-</td>""")
		out.append("""		<td align="right"><span class="sortkey">%u </span><a style="cursor:default" title="%s B">%s</a></td>""" % (qu.crealsize,decimal_number(qu.crealsize),humanize_number(qu.crealsize,"&nbsp;")))
		if qu.qflags&8:
			if qu.srealsize>0:
				if qu.srealsize>=qu.crealsize:
					clz="NOTEXCEEDED"
				elif qu.timetoblock>0:
					clz="SEXCEEDED"
				else:
					clz="EXCEEDED"
				out.append("""		<td align="right"><span class="%s">%.2f</span></td>""" % (clz,(100.0*qu.crealsize)/qu.srealsize))
			else:
				out.append("""		<td align="right"><span class="sortkey">%.2f </span><span class="EXCEEDED">inf</span></td>""" % (maxperc))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span>-</td>""")
		if qu.qflags&128:
			if qu.hrealsize>0:
				if qu.hrealsize>qu.crealsize:
					clz="NOTEXCEEDED"
				else:
					clz="EXCEEDED"
				out.append("""		<td align="right"><span class="%s">%.2f</span></td>""" % (clz,(100.0*qu.crealsize)/qu.hrealsize))
			else:
				out.append("""		<td align="right"><span class="sortkey">%.2f </span><span class="EXCEEDED">inf</span></td>""" % (maxperc))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span>-</td>""")
		out.append("""	</tr>""")

	out.append("""</table>""")
	return out


import sys
from common.constants import *
from common.utils import *
from common.utilsgui import *

def render(dp, fields, vld):
	out = []
	out.append("""<div id="footer">""")
	if dp.master()!=None:
		instancename=""
		if dp.master().has_feature(FEATURE_INSTANCE_NAME):
			data,length = dp.master().command(CLTOMA_INSTANCE_NAME,MATOCL_INSTANCE_NAME)
			if length>0:
				pos = 0
				err = 0
				instancename,pos,err = get_string_from_packet(data,pos,err)
		if instancename=="":
			instancename="(no name)"
	else:
		instancename="(unknown)"
	# Instance name
	out.append("""<div class="footer-left" data-tt="gui_instance_name"><span class="lg sm">Instance: </span><b>%s</b></div>""" % instancename)	
	
	# Auto refresh switch
	out.append("""<div class="footer-left" style="padding-left: 14px;" data-tt="gui_refresh_switch">""")
	out.append(""" <label class="switch" for="auto-refresh" id="auto-refresh-slide"><input type="checkbox" id="auto-refresh" />""")
	out.append(""" <span class="slider round"><span class="slider-text"></span></span></label>""")
	out.append("""</div>""")

	# Refresh button
	out.append("""<div class="lg sm md footer-left" style="padding-left: 3px; padding-right: 6px;">""")
	out.append(""" <span id="refresh-button" data-tt="gui_refresh_manual">&nbsp;Refresh:</span>&nbsp;<span id="refresh-timestamp" data-tt="gui_refresh_timestamp">%s</span>""" % datetime_to_str(time.time()))
	out.append("""</div>""")

	# GUI and python version
	out.append("""<div class="lg sm md footer-right text-icon">""")
	prover=""
	guiver = VERSION
	ci = dp.get_clusterinfo()
	if ci:
		issues=vld.check_gui_ver(ci.strver, guiver)
		data_help = issues.data_help()
		icon = issues.icon()
		text_class = issues.css_class()
	else:
		data_help=""
		icon=""
		text_class=""
	out.append("""<span class="lg sm">GUI&nbsp;</span><span style="padding-bottom:0;">%s</span><span class="%s" data-tt="%s">v.%s%s</span><span class="lg sm">&nbsp;&nbsp;Python v.%u.%u</span>""" % (icon, text_class, data_help, VERSION, prover, sys.version_info[0], sys.version_info[1]))
	out.append("""</div>""")

	# Dark mode switch
	out.append("""<div class="footer-right" id="theme-toggle" data-tt="gui_dark_mode"><svg height="14px" width="14px"><use xlink:href="#icon-light"/></svg></div>""")
	
	# Tooltip on/off switch
	out.append("""<div class="footer-right" data-tt="gui_tt_onoff">""")
	out.append(""" <label class="switch" for="tt-onoff" id="tt-onoff-slide"><input checked type="checkbox" id="tt-onoff"/>""")
	out.append(""" <span class="slider round"><span class="slider-text" style="font-size: 11.5px;">?</span></span></label>""")
	out.append("""</div>""")

	out.append("""</div>""")
	out.append("""  <script type="text/javascript" src="assets/acidtab.js"></script>""")
	out.append("""  <script type="text/javascript" src="assets/guiscripts.js"></script>""")
	out.append("""</body>""")
	out.append("""</html>""")

	return out


from common.utils import *
from common.utilsgui import *

def render(dp, fields, vld):
	out = []
	icharts = (
		(100,0,'cpu','cpu usage (percent)','<b>cpu usage</b>, sys: BOXGR2A user: BOXGR2B','sys: BOXGR2A user: BOXGR2B'),
		(101,0,'memory','memory usage (if available) rss + virt)','<b>memory usage</b> (if available), rss: BOXGR2A + virt: BOXGR2B','rss: BOXGR2A + virt: BOXGR2B'),
		(102,0,'space','raw disk space usage (used/total)','<b>raw disk space usage</b>, used: BOXGR2A total: BOXGR2B','used: BOXGR2A total: BOXGR2B'),
		(108,0,'objects','number of meta objects (others/files)','<b>number of meta objects</b>, others: BOXGR2A files: BOXGR2B','others: BOXGR2A files: BOXGR2B'),
		(109,0,'chunks','number of chunks (ec8/ec4/copy)','<b>number of chunks</b>, ec8: BOXGR3A ec4: BOXGR3B copy: BOXGR3C','ec8: BOXGR3A ec4: BOXGR3B copy: BOXGR3C'),
		(110,0,'regunder','number of regular chunks in danger','<b>number of regular chunks</b>, endangered: BOXGR2A undergoal: BOXGR2B','endangered: BOXGR2A undergoal: BOXGR2B'),
		(111,0,'allunder','number of all chunks in danger','<b>number of all chunks</b>, endangered: BOXGR2A undergoal: BOXGR2B ','endangered: BOXGR2A undergoal: BOXGR2B'),
		(112,0,'cservers','number of chunk servers','<b>number of chunk servers</b>, disconnected: BOXGR3A disconnected in maintenance: BOXGR3B working: BOXGR3C','disconnected: BOXGR3A disconnected in maintenance: BOXGR3B working: BOXGR3C'),
		(67,0,'udiff','space usage difference','<b>difference in space usage between the most and the least used chunk server</b>',''),
		(63,0,'delay','master max delay in seconds','<b>master max delay in seconds</b>',''),
		(103,0,'dels','chunk deletions per minute','<b>chunk deletions</b> per minute, unsuccessful: BOXGR2A successful: BOXGR2B','unsuccessful: BOXGR2A successful: BOXGR2B'),
		(104,0,'repl','chunk replications per minute','<b>chunk replications</b> per minute, unsuccessful: BOXGR2A successful: BOXGR2B','unsuccessful: BOXGR2A successful: BOXGR2B'),
		(105,0,'creat','chunk creations per minute','<b>chunk creations</b> per minute, unsuccessful: BOXGR2A successful: BOXGR2B','unsuccessful: BOXGR2A successful: BOXGR2B'),
		(106,0,'change','chunk internal operations per minute','<b>chunk internal operations</b> per minute, unsuccessful: BOXGR2A successful: BOXGR2B','unsuccessful: BOXGR2A successful: BOXGR2B'),
		(107,0,'split','chunk local split operations per minute','<b>chunk local split operations</b> per minute, unsuccessful: BOXGR2A successful: BOXGR2B','unsuccessful: BOXGR2A successful: BOXGR2B'),
		(68,0,'mountbytrcvd','traffic from cluster, data only (bytes per second)','',''),
		(69,0,'mountbytsent','traffic to cluster, data only (bytes per second)','',''),
		(49,0,'bread','traffic from cluster, data+overhead (bytes per second)','',''),
		(50,0,'bwrite','traffic to cluster, data+overhead (bytes per second)','',''),
		(21,0,'prcvd','packets received (per second)','',''),
		(22,0,'psent','packets sent (per second)','',''),
		(23,0,'brcvd','bits received (per second)','',''),
		(24,0,'bsent','bits sent (per second)','',''),
		(4,1,'statfs','statfs operations (per minute)','',''),
		(5,1,'getattr','getattr operations (per minute)','',''),
		(6,1,'setattr','setattr operations (per minute)','',''),
		(7,1,'lookup','lookup operations (per minute)','',''),
		(8,1,'mkdir','mkdir operations (per minute)','',''),
		(9,1,'rmdir','rmdir operations (per minute)','',''),
		(10,1,'symlink','symlink operations (per minute)','',''),
		(11,1,'readlink','readlink operations (per minute)','',''),
		(12,1,'mknod','mknod operations (per minute)','',''),
		(13,1,'unlink','unlink operations (per minute)','',''),
		(14,1,'rename','rename operations (per minute)','',''),
		(15,1,'link','link operations (per minute)','',''),
		(16,1,'readdir','readdir operations (per minute)','',''),
		(17,1,'open','open operations (per minute)','',''),
		(51,1,'read','read operations (per minute)','',''),
		(52,1,'write','write operations (per minute)','',''),
		(53,1,'fsync','fsync operations (per minute)','',''),
		(56,1,'truncate','truncate operations (per minute)','',''),
		(61,1,'create','file create operations (per minute)','',''),
		(54,1,'lock','file lock operations (per minute)','',''),
		(55,1,'snapshot','snapshot operations (per minute)','',''),
		(57,1,'getxattr','getxattr operations (per minute)','',''),
		(58,1,'setxattr','setxattr operations (per minute)','',''),
		(59,1,'getfacl','getfacl operations (per minute)','',''),
		(60,1,'setfacl','setfacl operations (per minute)','',''),
		(62,1,'meta','all meta data operations (per minute)','<b>all meta operations</b> - sclass, trashretention, eattr, etc. (per minute)','')
	)

	MCdata = fields.getstr("MCdata", "")

	charts = ([],[])
	for id,sheet,oname,desc,fdesc,sdesc in icharts:
		if not dp.master().is_pro() and ( id==63 ):
			continue
		if fdesc=='':
			fdesc = '<b>' + desc.replace(' (per', '</b> (per').replace(' (bytes per', '</b> (bytes per')
			if not '</b>' in fdesc:
				fdesc = fdesc + '</b>'
		fdeschtml = fdesc.replace('BOXGR1A','<span class="CBOX GR1A"></span>').replace('BOXGR2A','<span class="CBOX GR2A"></span>').replace('BOXGR2B','<span class="CBOX GR2B"></span>').replace('BOXGR3A','<span class="CBOX GR3A"></span>').replace('BOXGR3B','<span class="CBOX GR3B"></span>').replace('BOXGR3C','<span class="CBOX GR3C"></span>')
		sdeschtml = sdesc.replace('BOXGR1A','<span class="CBOX GR1A"></span>').replace('BOXGR2A','<span class="CBOX GR2A"></span>').replace('BOXGR2B','<span class="CBOX GR2B"></span>').replace('BOXGR3A','<span class="CBOX GR3A"></span>').replace('BOXGR3B','<span class="CBOX GR3B"></span>').replace('BOXGR3C','<span class="CBOX GR3C"></span>')
		charts[sheet].append((id,oname,desc,fdeschtml,sdeschtml))
	if MCdata=="" and dp.cluster.leaderfound():
		MCdata="%s:%u:%u" % (dp.master().host,dp.master().port,10 if dp.master().version_at_least(4,31,0) else 11)
	servers = []
	entrystr = []
	entrydesc = {}
	mservers = dp.get_masterservers(0) # sort by ip/port
	if len(mservers)>0:
		for sheet in [0,1]:
			for id,oname,desc,fdesc,sdesc in charts[sheet]:
				name = oname.replace(":","")
				entrystr.append(name)
				entrydesc[name] = desc
		for ms in mservers:
			if ms.version>=(2,0,15):
				chmode = 10 if ms.version>=(4,31,0) else 11
				name = "%s:%u" % (ms.strip,ms.port)
				namearg_res = "%s:%u" % (name,chmode)
				namearg_ops = "%s:%u" % (name,chmode+10)
				hostx = resolve(ms.strip,0,UNRESOLVED)
				if hostx==UNRESOLVED:
					host = "Server"
				else:
					host = hostx
				entrystr.append(namearg_res)
				entrystr.append(namearg_ops)
				entrydesc[namearg_res] = "%s: %s - resources %s" % (host, name," (leader)" if (dp.cluster.leaderfound() and ms.strip==dp.master().host) else "")
				entrydesc[namearg_ops] = "%s: %s - operations %s" % (host,name," (leader)" if (dp.cluster.leaderfound() and ms.strip==dp.master().host) else "")
				servers.append((ms.strip,ms.port,"ma_"+name.replace(".","_").replace(":","_"),"Server: <b>%s</b>" % (name),chmode))

	mchtmp = MCdata.split(":")
	if len(mchtmp)==2:
		mchtmp = (mchtmp[0],mchtmp[1],0)
	if len(mchtmp)==3:
		mahost = mchtmp[0]
		maport = mchtmp[1]
		mamode = int(mchtmp[2])
		if mamode>=20:
			mamode -= 20
			masheet = 1
		elif mamode>=10:
			mamode -= 10
			masheet = 0
		else:
			mamode = 1
			masheet = 0

		out.append("""<div class="tab_title">Master server charts</div>""")
		out.append("""<form action="#"><table class="FR">""")
		out.append("""	<tr>""")
		out.append("""		<th class="knob-cell chart-range">""")
		out.append(html_knob_selector_chart_range('ma_main'))
		out.append("""		</th>""")
		out.append("""		<th class="chart-select">""")
		out.append("""			<div class="select-fl"><select class="chart-select" name="madata" size="1" onchange="ma_change_data(this.selectedIndex,this.options[this.selectedIndex].value)">""")
		if MCdata not in entrystr:
			out.append("""				<option value="" selected="selected"> data type or server</option>""")
		for estr in entrystr:
			if estr==MCdata:
				out.append("""<option value="%s" selected="selected">%s</option>""" % (estr,entrydesc[estr]))
			else:
				out.append("""<option value="%s">%s</option>""" % (estr,entrydesc[estr]))
		out.append("""			</select><span class="arrow"></span></div>""")
		out.append("""		</th>""")
		out.append("""	</tr>""")

		if MCdata in entrystr:
			for id,name,desc,fdesc,sdesc in charts[masheet]:
				out.append("""	<tr class="C2 CHART">""")
				out.append("""		<td id="ma_%s_p" colspan="3" style="height:124px;" valign="middle">""" % name)
				out.append("""			<div class="CHARTJSW">""")
				out.append("""				<div id="ma_%s_c" class="CHARTJSC">""" % name)
				out.append("""					<span class="chartcaptionjs">%s</span>""" % fdesc)
				out.append("""				</div>""")
				out.append("""			</div>""")
				out.append("""			<div id="ma_%s_l" class="CHARTJSL"></div>""" % name)
				out.append("""			<div id="ma_%s_r" class="CHARTJSR"></div>""" % name)
				out.append("""		</td>""")
				out.append("""	</tr>""")
		out.append("""</table></form>""")

		if MCdata in entrystr:
			out.append("""<div class="tab_title">Master server charts (comparison)</div>""")
			out.append("""<form action="#"><table class="FR">""")
			for i in range(2):
				out.append("""	<tr>""")
				if i==0:
					out.append("""		<th class="knob-cell chart-cmp-range">""")
					out.append(html_knob_selector_chart_range('ma_cmp')) 
				else:
					out.append("""		<th style="text-align: right;vertical-align:middle;border:none;">""")
					out.append("""versus:""")
				out.append("""		</th>""")
				out.append("""		<th class="chart-cmp-select">""")
				out.append("""			<div class="select-fl"><select class="chart-select" id="machart%u_select" name="machart%u" size="1" onchange="ma_change_type(%u,this.options[this.selectedIndex].value)">""" % (i,i,i))
				no = 0
				for id,name,desc,fdesc,sdesc in charts[0]:
					out.append("""				<option value="%u">%s</option>""" % (no,desc))
					no += 1
				for id,name,desc,fdesc,sdesc in charts[1]:
					out.append("""				<option value="%u">%s</option>""" % (no,desc))
					no += 1
				out.append("""			</select><span class="arrow"></span></div>""")	
				out.append("""		</th>""")
				out.append("""	</tr>""")
				out.append("""	<tr class="C2 CHART">""")
				out.append("""		<td id="ma_%u_p" colspan="2" style="height: 124px; border-top: none;" valign="middle">""" % i)
				out.append("""			<div class="CHARTJSW">""")
				out.append("""				<div id="ma_%u_c" class="CHARTJSC">""" % i)
				out.append("""					<span id="ma_%u_d" class="chartcaptionjs">%s</span>""" % (i,charts[0][0][3]))
				out.append("""				</div>""")
				out.append("""			</div>""")
				out.append("""			<div id="ma_%u_l" class="CHARTJSL"></div>""" % i)
				out.append("""			<div id="ma_%u_r" class="CHARTJSR"></div>""" % i)
				out.append("""		</td>""")
				out.append("""	</tr>""")
			out.append("""</table></form>""")

			out.append("""<script type="text/javascript">""")
			out.append("""<!--//--><![CDATA[//><!--""")
			out.append("""	var ma_vids = [%s];""" % ",".join(map(repr,[ x[0] for x in charts[masheet] ])))
			out.append("""	var ma_inames = [%s];""" % ",".join(map(repr,[ x[1] for x in charts[masheet] ])))
			out.append("""	var ma_idesc = [%s];""" % ",".join(map(repr,[ x[3] for x in charts[masheet] ])))
			out.append("""	var ma_vids_cmp = [%s];""" % ",".join(map(repr,[ x[0] for x in (charts[0]+charts[1]) ])))
			out.append("""	var ma_idesc_cmp = [%s];""" % ",".join(map(repr,[ x[3] for x in (charts[0]+charts[1]) ])))
			out.append("""	var ma_hosts = [%s];""" % ",".join(map(repr,[ x[0] for x in servers ])))
			out.append("""	var ma_ports = [%s];""" % ",".join(map(repr,[ x[1] for x in servers ])))
			out.append("""	var ma_modes = [%s];""" % ",".join(map(repr,[ x[4]-10 for x in servers ])))
			out.append("""	var ma_host = "%s";""" % mahost)
			out.append("""	var ma_port = "%s";""" % maport)
			out.append("""	var ma_mode = %u;""" % mamode)
			out.append("""	var ma_valid = 1;""")
			out.append("""	var ma_base_href = "%s";""" % fields.createrawlink({"MCdata":""}))
			out.append("""//--><!]]>""")
			out.append("""</script>""")
		else:
			out.append("""<script type="text/javascript">""")
			out.append("""<!--//--><![CDATA[//><!--""")
			out.append("""	var ma_valid = 0;""")
			out.append("""	var ma_base_href = "%s";""" % fields.createrawlink({"MCdata":""}))
			out.append("""//--><!]]>""")
			out.append("""</script>""")
		out.append("""<script type="text/javascript">
<!--//--><![CDATA[//><!--
//#1 - comparison charts
function ma_create_charts(show_loading=true) {
ma_charttab = [];
var i;
if (ma_valid) {
for (i=0 ; i<ma_inames.length ; i++) {
	ma_charttab.push(new AcidChartWrapper("ma_"+ma_inames[i]+"_","ma_main",ma_host,ma_port,ma_mode,ma_vids[i],show_loading));
}
if (ma_chartcmp[0]) document.getElementById('machart0_select').value = ma_vids_cmp.indexOf(ma_chartcmp[0].id);
if (ma_chartcmp[1]) document.getElementById('machart1_select').value = ma_vids_cmp.indexOf(ma_chartcmp[1].id);
ma_chartcmp[0] = new AcidChartWrapper("ma_0_","ma_cmp",ma_host,ma_port,ma_mode,(ma_chartcmp[0]) ? ma_chartcmp[0].id : ma_vids_cmp[0],show_loading);
ma_chartcmp[1] = new AcidChartWrapper("ma_1_","ma_cmp",ma_host,ma_port,ma_mode,(ma_chartcmp[1]) ? ma_chartcmp[1].id : ma_vids_cmp[0],show_loading);
}
}

// main charts data update
function ma_change_data(indx, mcdata) {
var sindx,i;
if (ma_valid) {
sindx = indx - ma_inames.length;
if (sindx >= 0 && sindx < ma_modes.length) {
	for (i=0 ; i<ma_vids.length ; i++) {
		ma_charttab[i].set_host_port_id(ma_hosts[sindx],ma_ports[sindx],ma_modes[sindx],ma_vids[i]);
	}
}
}
rotateKnob("ma_main", -135);
document.location.replace(ma_base_href + "&MCdata=" + mcdata);
}

// compare charts data update
function ma_change_type(chartid, indx) {
var descel;
if (ma_valid) {
ma_chartcmp[chartid].set_id(ma_vids_cmp[indx]);
descel = document.getElementById("ma_"+chartid+"_d");
descel.innerHTML = ma_idesc_cmp[indx];
}
}

// charts initialization
ma_charttab = [];
ma_chartcmp = [];
ma_create_charts();
//--><!]]>
</script>""")

	elif len(mchtmp)==1 and len(MCdata)>0:
		chid = 0
		sdescadd = ''
		for chlist in charts:
			for id,name,desc,fdesc,sdesc in chlist:
				if name==MCdata:
					chid = id
					if sdesc!='':
						sdescadd = ', '+sdesc
		if chid==0:
			try:
				chid = int(MCdata)
			except Exception:
				pass
		if chid<=0 or chid>=1000:
			MCdata = ""

		out.append("""<div class="tab_title">Master server charts</div>""")
		out.append("""<form action="#"><table class="FR">""")
		out.append("""	<tr>""")
		out.append("""		<th class="knob-cell chart-range">""")
		out.append(html_knob_selector_chart_range('ma_main'))
		out.append("""		</th>""")
		out.append("""		<th class="chart-select">""")
		out.append("""			<div class="select-fl"><select class="chart-select" name="madata" size="1" onchange="ma_change_data(this.selectedIndex,this.options[this.selectedIndex].value)">""")
		if MCdata not in entrystr:
			out.append("""				<option value="" selected="selected"> data type or server</option>""")
		for estr in entrystr:
			if estr==MCdata:
				out.append("""<option value="%s" selected="selected">%s</option>""" % (estr,entrydesc[estr]))
			else:
				out.append("""<option value="%s">%s</option>""" % (estr,entrydesc[estr]))
		out.append("""			</select><span class="arrow"></span></div>""")
		if MCdata in entrystr:
			out.append("""		<label class="switch" for="ma_commonscale" id="ma_commonscale-slide" style="margin-left:10px;"><input type="checkbox" id="ma_commonscale"  onchange="AcidChartSetCommonScale('ma_main',this.checked)"/><span class="slider round"></span><span class="text">Use common Y-scale</span></label>""")				
		out.append("""		</th>""")
		out.append("""	</tr>""")
		if MCdata in entrystr:
			for mahost,maport,name,desc,chmode in servers:
				out.append("""	<tr class="C2 CHART">""")
				out.append("""		<td id="ma_%s_p" colspan="3" style="height:124px;" valign="middle">""" % name)
				out.append("""			<div class="CHARTJSW">""")
				out.append("""				<div id="ma_%s_c" class="CHARTJSC">""" % name)
				out.append("""					<span id="ma_%s_d" class="chartcaptionjs">%s%s</span>""" % (name,desc,sdescadd))
				out.append("""				</div>""")
				out.append("""			</div>""")
				out.append("""			<div id="ma_%s_l" class="CHARTJSL"></div>""" % name)
				out.append("""			<div id="ma_%s_r" class="CHARTJSR"></div>""" % name)
				out.append("""		</td>""")
				out.append("""	</tr>""")
		out.append("""</table></form>""")

		if MCdata in entrystr:
			out.append("""<script type="text/javascript">""")
			out.append("""<!--//--><![CDATA[//><!--""")
			out.append("""	var ma_vids = [%s];""" % ",".join(map(repr,[ x[0] for x in (charts[0]+charts[1]) ])))
			out.append("""	var ma_idesc = [%s];""" % ",".join(map(repr,[ x[4] for x in (charts[0]+charts[1]) ])))
			out.append("""	var ma_hosts = [%s];""" % ",".join(map(repr,[ x[0] for x in servers ])))
			out.append("""	var ma_ports = [%s];""" % ",".join(map(repr,[ x[1] for x in servers ])))
			out.append("""	var ma_inames = [%s];""" % ",".join(map(repr,[ x[2] for x in servers ])))
			out.append("""	var ma_sdesc = [%s];""" % ",".join(map(repr,[ x[3] for x in servers ])))
			out.append("""	var ma_modes = [%s];""" % ",".join(map(repr,[ x[4]-10 for x in servers ])))
			out.append("""	var ma_chid = %u;""" % chid)
			out.append("""	var ma_valid = 1;""")
			out.append("""	var ma_base_href = "%s";""" % fields.createrawlink({"MCdata":""}))
			out.append("""//--><!]]>""")
			out.append("""</script>""")
		else:
			out.append("""<script type="text/javascript">""")
			out.append("""<!--//--><![CDATA[//><!--""")
			out.append("""	var ma_valid = 0;""")
			out.append("""	var ma_base_href = "%s";""" % fields.createrawlink({"MCdata":""}))
			out.append("""//--><!]]>""")
			out.append("""</script>""")

		out.append("""<script type="text/javascript">
<!--//--><![CDATA[//><!--
//#2 - no comparison charts
function ma_create_charts(show_loading=true) {
var i;
ma_charttab = [];
if (ma_valid) {
for (i=0 ; i<ma_inames.length ; i++) {
	ma_charttab.push(new AcidChartWrapper("ma_"+ma_inames[i]+"_","ma_main",ma_hosts[i],ma_ports[i],ma_modes[i],ma_chid,show_loading));
}
}
}

// main charts data update
function ma_change_data(indx, mcdata) {
var i,chartid;
var descel;
if (ma_valid && indx >= 0 && indx < ma_vids.length) {
for (i=0 ; i<ma_inames.length ; i++) {
	chartid = ma_inames[i];
	ma_charttab[i].set_id(ma_vids[indx]);
	descel = document.getElementById("ma_"+chartid+"_d");
	if (ma_idesc[indx]!="") {
		descel.innerHTML = ma_sdesc[i] + " " + ma_idesc[indx];
	} else {
		descel.innerHTML = ma_sdesc[i];
	}
}
}
document.location.replace(ma_base_href + "&MCdata=" + mcdata);
}

// charts initialization
ma_charttab = [];
ma_create_charts();
//--><!]]>
</script>""")
	else:
		out.append("""<div class="tab_title">Master server charts</div>""")
		out.append("""<form action="#"><table class="FR">""")
		out.append("""	<tr>""")
		out.append("""		<th>""")
		out.append("""			Select: <div class="select-fl"><select class="chart-select" name="madata" size="1" onchange="document.location.href='%s&MCdata='+this.options[this.selectedIndex].value">""" % fields.createrawlink({"MCdata":""}))
		if MCdata not in entrystr:
			out.append("""				<option value="" selected="selected"> data type or server</option>""")
		for estr in entrystr:
			if estr==MCdata:
				out.append("""<option value="%s" selected="selected">%s</option>""" % (estr,entrydesc[estr]))
			else:
				out.append("""<option value="%s">%s</option>""" % (estr,entrydesc[estr]))
		out.append("""			</select><span class="arrow"></span></div>""")
		out.append("""		</th>""")
		out.append("""	</tr>""")
		out.append("""</table></form>""")

	return out


import math

from common.constants import *
from common.utils import *
from common.utilsgui import *
from common.models import *

# Capacity etc. Warning/Error thresholds
thrsh_wrn = THRESHOLD_WARNING
thrsh_err = THRESHOLD_ERROR

padding = 0       # graph padding (all borders)
tile_padding = 10 # tile padding  

ms_top = 10            # master servers column top position
ms_left = tile_padding # master servers column left position
ms_width = 260         # master server icon width
ms_height = 50         # master server icon height
ms_margin_top = 30     # master servers vertical spacing 

mls_tile_margin_top = 20 # metaloggers column top margin
mls_left = tile_padding # metaloggers column left position
mls_width = ms_width    # metaloggers icon width
mls_height = 30         # metaloggers icon height
mls_margin_top = 30     # metaloggers vertical spacing

cs_top = ms_top    # chunk servers column top position
cs_left_1st = 490  # chunk servers 1st column left position
cs_left_2nd = 800  # chunk servers 2nd column left position
cs_width = ms_width     # chunk server icon width
cs_height = 50     # chunk server icon height
cs_margin_top = 40 # chunk servers vertical spacing 

net_backbone_hmargin = 12
net_backbone_x = ms_width+130

green =  'fill: var(--ok-clr);'
orange = 'fill: var(--warning-clr);'
red =    'fill: var(--error-clr);'
bold =   'font-weight: bold;'
left =   'text-anchor:start;'
middle = 'text-anchor:middle;'
right =  'text-anchor:end;'
em08 =  'font-size:0.8em;'
em09 =  'font-size:0.9em;'
em11 =  'font-size:1.1em;'
em12 =  'font-size:1.2em;'
em14 =  'font-size:1.4em;'

# def defs():
  # out = []
  # out.append('<defs>')
  # out.append('</defs>')
  # return "\n".join(out)

# def styles():
#   out = []
#   out.append('<style>')
#   out.append('</style>')
#   return "\n".join(out)

# Returns x.y.z version number (without "PRO" suffix)
def version_number(strver):
  if strver.find(' PRO') > 0:
    return strver[:-4]
  else:
    return strver

def humanize(number, thousand=1000, suff=''):
  if number==None: return 'n/a'
  number*=100
  scale=0
  while number>=99950:
    number = number//thousand
    scale+=1
  if number<995 and scale>0:
    b = (number+5)//10
    nstr = "%u.%u" % divmod(b,10)
  else:
    b = (number+50)//100
    if number>0 and b==0:
      nstr = "<&#8239;1" # number is small but not zero
    else:
      nstr = "%u" % b
  if scale>0:
    if (suff=='iB' or suff=='iB/s'):
      slist="-kMGTPEZY"
    else:
      slist=['','k','M',' Bn']
    return (nstr,"%s%s" % (slist[scale],suff))
  else:
    suff = 'B' if suff=='iB' else suff
    suff = 'B/s' if suff=='iB/s' else suff
    return (nstr,suff)

def humanize_str(number, thousand=1000, suff=''):
  if number==None: return 'n/a'
  (str, suff) = humanize(number, thousand, suff)
  return str+'&#8239;'+suff

def humanize_bytes(number, suff='iB'):
  if number==None: return 'n/a'
  return humanize(number, 1024, suff)

def humanize_bytes_str(number, suff='iB'):
  if number==None: return 'n/a'
  (str, suff) = humanize_bytes(number, suff)
  return str+'&#8239;'+suff

# Translates given percent (0..100) number into approximate text
def percent2approx(percent):
  if percent==0.0:
    return '0'
  if percent<0.1:
    return '<1&permil;'
  if percent<1:
    return '<1%'
  if percent<99:
    return '%.0f%%' % percent
  if percent>=100:
    return 'all'
  return '>%.0f%%' % percent

# Translates given tag to x,y by adding the "g transform" tag
def translate(tag, x, y, onclick=''):
  if onclick:
    return '<g transform="matrix(1,0,0,1,%s,%s)" style="cursor: pointer;" onclick="%s">\n %s\n</g>' % (x, y, onclick, tag)
  else:
    return '<g transform="matrix(1,0,0,1,%s,%s)">\n %s\n</g>' % (x, y, tag)

# Generate text with class clz at position x, y
def text(msg, x, y, clz, maxLength=0, style=''):
  if maxLength>0:
    if len(msg) > maxLength:
      if msg[maxLength-2] == '.':
        msg = (msg[:maxLength-2])
      else:
        msg = (msg[:maxLength-1])
      msg += '...' #'…'
  return '<text class="%s" style="%s" x="%s" y="%s">%s</text>' % (clz, style, x, y, msg)

# Generates arc at x, y (left-upper corner of a surrounding box) with given radius starting and ending at given angles: <path fill="none" stroke="#446688" stroke-width="3" d="M 184.64101615137756 170 A 40 40 0 1 0 115.35898384862246 170"></path>
def arc(x, y, radius, startAngle, endAngle, strokeWidth=8, clz=''):
  def  polarToCartesian(x, y, radius, angleInDegrees):
    angleInRadians = (angleInDegrees-90) * 0.0174532 #math.pi / 180.0
    return (x + (radius * math.cos(angleInRadians)),y + (radius * math.sin(angleInRadians)))
  x=x+radius+(strokeWidth/2.0)
  y=y+radius+(strokeWidth/2.0)
  (startx, starty) = polarToCartesian(x, y, radius, endAngle)
  (endx, endy) = polarToCartesian(x, y, radius, startAngle)
  largeArcFlag = '0' if endAngle - startAngle <= 180 else '1';
  points = 'M %f %f A %f %f 0 %s 0 %f %f' % (startx, starty, radius, radius, largeArcFlag, endx, endy)
  return'<path class="%s" style="stroke-width: %fpx" d="%s"/>' % (clz,strokeWidth,points)

# Generate arc gauge with name, green, orange or red depending on val threshold (75%/90%) 
def arc_gauge(x, y, name, label, val, label_style='', r=12, arcTo=120, strokeWidth=8): 
  out = []
  if name:
    out.append(text(name, x+r+strokeWidth/2.0, y, 'arc-gauge-name', 0, label_style))
  out.append(arc(x, y+strokeWidth/2.0, r, -arcTo, arcTo, strokeWidth, 'arc-gauge-bckgrd'))
  val = min(val, 1.0)
  val = max(val, 0.0)
  clz = 'arc-gauge-green' if val < thrsh_wrn else ('arc-gauge-orange' if val < thrsh_err else 'arc-gauge-red')
  endAngle = val*2.0*arcTo - arcTo
  endAngle = max(endAngle, -arcTo+1)
  out.append(arc(x, y+strokeWidth/2.0, r, -arcTo, endAngle, strokeWidth, clz))
  if label:
    if (type(label) is tuple):
      out.append(text(label[0], x+r+strokeWidth/2, y+r+strokeWidth+2, 'val-label-value',0,middle+em11))
      out.append(text(label[1], x+r+strokeWidth/2, y+r+strokeWidth+12, 'val-label-units',0,middle+em08))
    else:
      out.append(text(label, x, y+13, 'arc-gauge-label'))
  return "\n".join(out)

# # Generate horizontal linear % gauge with name, green, orange or red depending on val threshold (75%/90%) 
# def hgauge(name, val, x, y):
#   out = []
#   out.append(text(name, x, y, 'label'))
#   h = 5
#   w = ms_width/3
#   out.append('<rect x="%f" y="%f" width="%f" height="%f" rx="1" ry="1" class="gauge-bckgrd"/>' % (x+3, y-h, w, h))
#   val = min(val, 1)
#   val = max(val, 0)
#   clz = 'gauge-green' if val < thrsh_wrn else ('gauge-orange' if val < thrsh_err else 'gauge-red')
#   w = val*w
#   w = max(w, 0.5)
#   out.append('<rect x="%f" y="%f" width="%f" height="%f" rx="1" ry="1" class="%s"/>' % (x+3, y-h, w, h, clz))
#   return "\n".join(out)

# Generates svg text memory gauge
def svg_mem_gauge(name, val, x, y):
  out=[]
  out.append(text(name, x, y, 'val-label-name'))
  (hval, units)=humanize_bytes(val)
  out.append(text(hval, x, y+15, 'val-label-value'))
  out.append(text(units, x, y+28, 'val-label-units'))
  return "\n".join(out)

# Generates div text memory gauge
def html_mem_gauge(name, val, color_class=''):
  (hval, units)=humanize_bytes(val)
  out=[]
  out.append('<table class="gauge">')
  if name:
    out.append('<tr><td class="name">%s</td></tr>' % name)
  out.append('<tr><td class="txt-value %s">%s</td></tr>' % (color_class,hval))
  out.append('<tr><td class="units %s">%s</td></tr>' % (color_class, units))
  out.append('</table>')
  return "\n".join(out)

def svg_tag(w, h, svg_body):
  return '<svg viewbox="0 0 %u %u" width="%u" height="%u" xmlns="http://www.w3.org/2000/svg">%s</svg>' % (w, h, w, h, svg_body)

# Generates a full span+svg tag with an icon
def html_icon(id, scale=1.0):
  return '<span class="icon"><svg height="%fpx" width="%fpx"><use transform="scale(%f)" xlink:href="#%s"/></svg></span>' % (12*scale, 12*scale, scale, id)

##############################################
# Generate cluster state html
def html_cluster_state(info):
  out = []

  out.append('<div class="info-cards">')
  if 'totalspace' in info and 'availspace' in info and info['totalspace']!=0:
    # Total and available size
    out.append('<div class="card pointer" onclick="showGraphInfo(\'IG\')" data-tt="st_card_cluster_space">')
    out.append('<table class="info-table">')
    out.append('<tr class="low"><td class="bold">Total space</td><td>Free</td></tr>')
    out.append('<tr><td>%s</td>' % svg_tag(2*20+8,2*20+8, arc_gauge(0,0, '', humanize_bytes(info['totalspace']), float(info['totalspace']-info['availspace'])/float(info['totalspace']), bold, 20, 120, 8))) 
    out.append('<td>%s</td></tr>' % html_mem_gauge('', info['availspace']))
    out.append('</table>')
    out.append('</div>') #card Total and available size
    out.append('<div class="card-divider"></div>')

  #Health
  severity=0
  if 'chunks_progress' in info and info['chunks_progress']==0:
    if 'chunks_total' in info and info['chunks_total']>0 and 'chunks_missing' in info and info['chunks_missing']>0:
      severity=2
      data_health='<td class="label-value"><span class="error-txt">%s missing</span></td></tr>' % percent2approx(100.0*info['chunks_missing']/float(info['chunks_total']))
    elif 'chunks_total' in info and info['chunks_total']>0 and 'chunks_undergoal' in info and 'chunks_endangered' in info and (info['chunks_undergoal']+info['chunks_endangered'])>0:
      severity=1
      data_health='<td class="label-value"><span class="warning-txt">%s in danger</span></td></tr>' % percent2approx(100.0*(info['chunks_undergoal']+info['chunks_endangered'])/float(info['chunks_total']))
    else:
      data_health='<td class="label-value">Normal</td></tr>'
  else:
    data_health='<td class="label-value"><span class="warning-txt">checking…</span></td></tr>'
  if 'servers_warnings' in info and len(info['servers_warnings'])>0:
    severity = 1 if severity < 2 else severity
    servers_health='<td class="label-value"><span class="warning-txt">%s abnormal</span></td></tr>' % len(info['servers_warnings'])
  else:
    servers_health='<td class="label-value">Normal</td></tr>'
  if severity==2:
    icon=html_icon('icon-error', 1.2)
  elif severity==1:
    icon=html_icon('icon-warning', 1.2)
  else:
    icon=html_icon('icon-ok-circle', 1.2)
  out.append('<div class="card">')
  out.append('<table class="info-table">')
  out.append('<tr><td colspan="2" class="card-title"><div class="text-icon" data-tt="st_card_health">Health %s</div></td></tr>' % icon)
  if 'leaderfound' in info and info['leaderfound']==1:
    out.append('<tr class="pointer" onclick="showGraphInfo(\'IC\',\'ICsclassid=-1\')"><td class="label-name">Data:</td>')
    out.append(data_health)
  else:
    out.append('<tr class="pointer" onclick="showGraphInfo(\'IG\')"><td class="label-name">Leader:</td>')
    out.append('<td class="label-value"><span class="error-txt">missing</span></td></tr>')
  out.append('<tr><td class="label-name">Servers:</td>')
  out.append(servers_health)
  out.append('</table>')
  out.append('</div>') #card health
  out.append('<div class="card-divider"></div>')

  #Usage
  out.append('<div class="card">')
  out.append('<table class="info-table">')
  out.append('<tr><td colspan="4" class="card-title" data-tt="st_card_usage">Usage</td></tr>') 
  out.append('<tr class="pointer"">')
  out.append('<td class="label-name" onclick="showGraphInfo(\'IG\')">Files:</td>')
  if 'files' in info and 'trfiles' in info:
    out.append('<td class="label-value" onclick="showGraphInfo( \'IG\')">%s</td>' % humanize_str(info['files']-info['trfiles']))
  else:
    out.append('<td class="label-value" onclick="showGraphInfo( \'IG\')">Unknown</td>')
  out.append('<td class="label-name" onclick="showGraphInfo(\'MS\')">Mounts:</td>')
  if 'mounts' in info:
    out.append('<td class="label-value" onclick="showGraphInfo(\'MS\')">%s</td>' % humanize_str(info['mounts']))
  else:
    out.append('<td class="label-value" onclick="showGraphInfo(\'MS\')">Unknown</td>')
  out.append('</tr>')
  out.append('<tr class="pointer"">')
  out.append('<td class="label-name" onclick="showGraphInfo( \'IG\')">Folders:</td>')
  if 'dirs' in info:
    out.append('<td class="label-value" onclick="showGraphInfo( \'IG\')">%s</td>' % humanize_str(info['dirs']))
  else:
    out.append('<td class="label-value" onclick="showGraphInfo( \'IG\'")>Unknown</td>')
  out.append('<td class="label-name" onclick="showGraphInfo( \'IG\')">Trash size:</td>')
  if 'trspace' in info :
    out.append('<td class="label-value" onclick="showGraphInfo( \'IG\')">%s</td>' % humanize_bytes_str(info['trspace']))
  else:
    out.append('<td class="label-value" onclick="showGraphInfo( \'IG\')">Unknown</td>')
  out.append('</tr>')
  out.append('</table>')
  out.append('</div>') #card usage

  out.append('<div class="card-divider"></div>')

  #Performance
  out.append('<div class="card" data-tt="st_card_performance">')
  out.append('<table class="info-table pointer" onclick="showGraphInfo(\'MO\')">')
  out.append('<tr><td colspan="4" class="card-title">Performance</td></tr>') 
  out.append('<tr>')
  if 'bread_ext' in info and info['bread_ext']!=None:
    out.append('<td class="label-name"">Outbound:</td>')
    out.append('<td class="label-value">%s</td>' % humanize_bytes_str(info['bread_ext'],'iB/s'))
  else:
    out.append('<td></td>')
    out.append('<td></td>')
  out.append('<td class="label-name">R/W ops:</td>')
  if 'rops' in info and 'wops' in info and info['rops']!=None and info['wops']!=None:
    out.append('<td class="label-value">%s/s</td>' % humanize_str(info['rops']+info['wops']))
  else:
    out.append('<td class="label-value">n/a</td>')
  out.append('</tr>')
  out.append('<tr>')
  if 'bwrite_ext' in info and info['bwrite_ext']!=None:
    out.append('<td class="label-name">Inbound:</td>')
    out.append('<td class="label-value">%s</td>' % humanize_bytes_str(info['bwrite_ext'],'iB/s'))
  else:
    out.append('<td></td>')
    out.append('<td></td>')
  out.append('<td class="label-name">All ops:</td>')
  if 'ops' in info and info['ops']!=None:
    out.append('<td class="label-value">%s/s</td>' % humanize_str(info['ops']))
  else:
    out.append('<td class="label-value">n/a</td>')
  out.append('</tr>')
  out.append('</table>')
  out.append('</div>') #card performance


  #Licence (if applicable)
  if 'licmaxsize' in info and 'currentsize' in info and info['licmaxsize']!=0  and info['currentsize']!=None and info['licmaxsize']!=None:
    licleft=info['licmaxsize']-info['currentsize']
    out.append('<div class="card-divider"></div>')
    out.append('<div class="card pointer" onclick="showGraphInfo(\'LI\')" data-tt="st_card_licence">')
    out.append('<table class="info-table">')
    if licleft < 0:
      liclefttxt = 'Over'
      licleftclass = 'ERROR'
    else:
      liclefttxt = 'Left'
      licleftclass = ''
    out.append('<tr class="low"><td class="bold">Licence</td><td>%s</td></tr>' % liclefttxt)
    out.append('<tr><td>%s</td>' % svg_tag(2*20+8,2*20+8,arc_gauge(0,0, '', humanize_bytes(info['licmaxsize']), (float(info['currentsize'])/float(info['licmaxsize'])), bold, 20, 120, 8))) 
    out.append('<td>%s</td></tr>' % html_mem_gauge('', abs(licleft), licleftclass))
    out.append('</table>')
    out.append('</div>') #card Licence

  # out.append('<div class="cards-underline"></div>')  
  out.append('</div> <!-- info-cards -->') #info-cards

  return "\n".join(out)

##############################################
# Generate master server icon 
def mserver(srv, net_offset_y, leader_strver):
  #TODO: add metadata in-sync / out-of-sync warnings
  live = 'live' in srv and srv['live']==1
  led_cls = 'led-normal'
  srv_tt = []
  out = []
  out.append('<rect width="%d" height="%d" rx="4" ry="4" class="srv-rect-bckgrnd"/>' % (ms_width, ms_height))
  out.append(text(srv['name'], 21, 20, "srv-name", 17))
  if not live:
    led_cls = 'led-error'
    srv_tt.append('follower_unreachable')
    out.append(text('UNREACHABLE', 7, 35, 'state-error'))
  if live:
    if 'strver' in srv and srv['strver'] != '0.0.0':
      ver_cls = "text-version"
      if leader_strver!='' and leader_strver!=srv['strver']:
        ver_cls += "-mismatch"  
        led_cls = 'led-warning'
      out.append('<line class="srv-line-ver" x1="%f" y1="%f" x2="%f" y2="%f"/>' % (ms_width-17, 0, ms_width-17, ms_height))
      out.append('<g transform="translate(%f,%f) rotate(90)">' % (ms_width-13, ms_height/2))
      out.append(text(version_number(srv['strver']), 0, 0, ver_cls))
      out.append('</g>')
    if 'statestr' in srv:
      if len(srv['statestr'])>15:
        words=srv['statestr'].split(' ')
        line1=" ".join(words[:2])
        line2=" ".join(words[2:])
        out.append(text(line1, 7, 34, 'srv-state-small'))
        out.append(text(line2, 7, 46, 'srv-state-small'))
      else:
        out.append(text(srv['statestr'], 7, 40, 'srv-state'))
      if not srv['statestr'] in ['LEADER', 'FOLLOWER', 'MASTER', '-']:
        led_cls = 'led-warning'
    if 'cpuload' in srv:
      # out.append(hgauge('CPU', srv['cpuload'], 28, 42))
      x=cs_width-100
      out.append('<line class="srv-line" x1="%f" y1="%f" x2="%f" y2="%f"/>' % (x, 7, x, ms_height-7))
      out.append(arc_gauge(x+5, 14, 'CPU', '', srv['cpuload']))
      if srv['cpuload'] > thrsh_wrn:
        led_cls = 'led-warning'
    if 'memory' in srv:
      x=ms_width-58
      out.append('<line class="srv-line" x1="%f" y1="%f" x2="%f" y2="%f"/>' % (x, 7, x, ms_height-7))
      out.append(text('RAM', x+20, 14, 'val-label-name'))
      (memval, memunits)=humanize_bytes(srv['memory'])
      out.append(text(memval, x+20, 29, 'val-label-value'))
      out.append(text(memunits, x+20, 42, 'val-label-units'))
  if 'ip' in srv and 'port' in srv:
    out.append(text(srv['ip'], ms_width+5, net_offset_y-3, 'ip-address-ms'))
    out.append(text(':'+str(srv['port']), ms_width+5, net_offset_y+12, 'ip-port-ms'))
  out.append('<circle cx="12" cy="14" r="5" class="%s"/>' % led_cls)
  out.append('<rect width="%d" height="%d" rx="4" ry="4" class="srv-rect"/>' % (ms_width, ms_height))
  if srv_tt:
    # wrap with <g> to display tooltip over entire rect area
    out.append('</g>')
    out.insert(0, '<g data-tt="%s">' % ",".join(srv_tt))
  return "\n".join(out)

##############################################
# Generate metalogger server icon
def mlserver(srv, net_offset_y, leader_strver):
  led_cls = 'led-normal'
  out = []
  out.append('<rect width="%d" height="%d" rx="4" ry="4" class="srv-rect-bckgrnd"/>' % (mls_width, mls_height))
  out.append(text(srv['name'], 21, 20, "srv-name", 24))
  if 'strver' in srv and srv['strver'] != '0.0.0':
    ver_cls = "text-version"
    if leader_strver!='' and leader_strver!=srv['strver']:
      ver_cls += "-mismatch"  
      led_cls = 'led-warning'
    out.append('<line class="srv-line-ver" x1="%f" y1="%f" x2="%f" y2="%f"/>' % (mls_width-40, 0, mls_width-40, mls_height))
    out.append(text(version_number(srv['strver']), mls_width-20, mls_height/2+5, ver_cls))
  if 'ip' in srv :
    out.append(text(srv['ip'], ms_width+5, net_offset_y-3, 'ip-address-ms'))
  out.append('<circle cx="12" cy="14" r="5" class="%s"/>' % led_cls)
  out.append('<rect width="%d" height="%d" rx="4" ry="4" class="srv-rect"/>' % (mls_width, mls_height))
  return "\n".join(out)

##############################################
# Generate chunk server icon
def cserver(srv, net_offset_y, leader_strver):
  live = 'live' in srv and srv['live']==1
  maintenance = 'maintenance' in srv and srv['maintenance']==1
  led_cls = 'led-normal'
  srv_tt = []
  out = []
  out.append('<rect width="%d" height="%d" rx="4" ry="4" class="srv-rect-bckgrnd"/>' % (ms_width, ms_height))
  out.append(text(srv['name'], 21, 20, "srv-name", 12))
  if not live:
    if maintenance:
      led_cls = 'led-maintenance'
      srv_tt.append('cs_unreachable_maintain')
      out.append(text('MAINTENANCE', 7, 35, "state-maintenance"))
      out.append(text('UNREACHABLE', 94, 35, "state-error"))
    else:
      led_cls = 'led-error'
      srv_tt.append('cs_unreachable')
      out.append(text('UNREACHABLE', 7, 35, "state-error"))
  
  if live:
    if maintenance:
      led_cls = 'led-maintenance'
      srv_tt.append('cs_maintain')
      out.append(text('MAINTENANCE', 7, 35, "state-maintenance"))
    if 'strver' in srv and srv['strver'] != '0.0.0':
      ver_cls = "text-version"
      if leader_strver!='' and leader_strver!=srv['strver']:
        ver_cls += "-mismatch"  
        led_cls = 'led-warning'
      out.append('<line class="srv-line-ver" x1="%f" y1="%f" x2="%f" y2="%f"/>' % (cs_width-17, 0, cs_width-17, cs_height))
      out.append('<g transform="translate(%f,%f) rotate(90)">' % (cs_width-13, cs_height/2))
      out.append(text(version_number(srv['strver']), 0, 0, ver_cls))
      out.append('</g>')
    if not maintenance and 'queue_state_msg' in srv:
      out.append(text('Queue:', 7, 33, 'label',0,left))
      if srv['queue_state_str'] == 'Overloaded': #FIXME: use queue_state and a numeric constant
        state_cls = 'state-error'
        led_cls = 'led-warning'
      else:
        state_cls = 'state-normal'
      out.append(text(srv['queue_state_msg'], 7, 44, state_cls))

    if 'hdd_reg_total' in srv and 'hdd_reg_used' in srv:
      x=cs_width-132
      out.append('<line class="srv-line" x1="%f" y1="%f" x2="%f" y2="%f"/>' % (x, 7, x, cs_height-7))
      if srv['hdd_reg_total']!=0:
        out.append(arc_gauge(x+6, 14, 'HDD', '', float(srv['hdd_reg_used'])/float(srv['hdd_reg_total']), bold))
        if srv['hdd_reg_used']/srv['hdd_reg_total'] > thrsh_wrn:
          led_cls = 'led-warning'
      else:
        out.append(text('HDD', x+22, 14, 'arc-gauge-name', 0, bold+middle))
      x+=36
      # out.append('<line class="srv-line" x1="%f" y1="%f" x2="%f" y2="%f"/>' % (x, 18, x, cs_height-7))
      out.append(text('Free', x+20, 14, 'val-label-name'))
      (hddfree, hddunits)=humanize_bytes(srv['hdd_reg_total']-srv['hdd_reg_used'])
      out.append(text(hddfree, x+20, 29, 'val-label-value'))
      out.append(text(hddunits, x+20, 42, 'val-label-units'))
      x+=36
      out.append(text('Health', x+20, 14, 'val-label-name'))
      if 'hdds_status' in srv:
        if srv['hdds_status'].startswith('error'):
          srv_tt.append('hdd_damage');
          out.append('<use xlink:href="#icon-error" x="%f" y="19" style="%s"/>' % (x+13, red))
          out.append(text('Errors', x+20, 42, 'value',0,middle+red+em08))
          led_cls = 'led-warning'
        elif srv['hdds_status'].startswith('warning'):
          srv_tt.append('hdd_error');
          out.append('<use xlink:href="#icon-warning" x="%f" y="19" style="%s"/>' % (x+13, orange))
          out.append(text('Warnings', x+20, 42, 'value',0,middle+orange+em08))
          led_cls = 'led-warning'
        elif srv['hdds_status'].startswith('ok'):
          out.append('<g transform="translate(%f,21)"><use xlink:href="#icon-ok-circle" transform="scale(1.5)"/></g>' % (x+11))
        else:
          out.append(text('unknown', x+20, 42, 'value',0,middle))
      else:
        out.append(text('unknown', x+20, 42, 'value',0,middle))

  if 'ip' in srv and 'port' in srv:
    out.append(text(srv['ip'], -5, net_offset_y-3, 'ip-address-cs'))
    out.append(text(":"+str(srv['port']), -5, net_offset_y+12, 'ip-port-cs'))

  out.append('<circle cx="12" cy="14" r="5" class="%s"/>' % led_cls)
  out.append('<rect width="%d" height="%d" rx="4" ry="4" class="srv-rect"/>' % (ms_width, ms_height))
  out.append('<rect width="%d" height="%d" rx="4" ry="4" class="srv-rect"/>' % (ms_width, ms_height))
  if srv_tt:
    # wrap with <g> to display tooltip over entire rect area
    out.append('</g>')
    out.insert(0, '<g data-tt="%s">' % ",".join(srv_tt))
  return "\n".join(out)

##############################################
# Generate SVG picture of a cluster
##############################################
def svg_mfs_graph( masterservers, metaloggers, chunkservers, leader_strver):
  svg_h = 0
  svg_w = net_backbone_x

  out = []  
  # out.append(defs())
  # out.append(styles())

  # Print master servers
  net_offset_y = 0.6 * ms_height
  x = ms_left
  y = ms_top
  net_start_y = y
  net_end_y = 0

  if len(masterservers) > 0:
    # master servers tile
    tile_width=(net_backbone_x-net_backbone_hmargin)-(x-tile_padding)
    tile_height=2*tile_padding+(len(masterservers)-1)*(ms_height+ms_margin_top)+ms_height+22
    out.append('<rect x="%f" y="%f" width="%f" height="%f" rx="4" ry="4" class="tile-bckgrnd"/>' % (x-tile_padding, y-tile_padding, tile_width, tile_height))
    out.append(text('Master servers', x-tile_padding+tile_width-8, y-tile_padding+tile_height-9, 'tile-title-right'))
    
    for ms in masterservers:
      out.append('<line class="network-line" x1="%f" y1="%f" x2="%f" y2="%f"/>' % (x+ms_width, y+net_offset_y, net_backbone_x, y+net_offset_y))
      out.append(translate(mserver(ms,net_offset_y,leader_strver), x, y, "showGraphInfo('IM')"))
      net_end_y = y+net_offset_y
      y+=ms_height+ms_margin_top
    y=ms_top+tile_height
    svg_h=max(svg_h, y+ms_height)

  # Print metaloggers
  if len(metaloggers) > 0:
    net_offset_y = 0.7 * mls_height
    x=mls_left
    y=y+mls_tile_margin_top
    # metaloggers tile
    tile_width=(net_backbone_x-net_backbone_hmargin)-(x-tile_padding)
    tile_height=2*tile_padding+(len(metaloggers)-1)*(mls_height+mls_margin_top)+mls_height+22
    out.append('<rect x="%f" y="%f" width="%f" height="%f" rx="4" ry="4" class="tile-bckgrnd"/>' % (x-tile_padding, y-tile_padding, tile_width, tile_height))
    out.append(text('Metaloggers', x-tile_padding+tile_width-8, y-tile_padding+tile_height-9, 'tile-title-right'))
    svg_h=max(svg_h, y-tile_padding+tile_height)

    for mls in metaloggers:
      out.append('<line class="network-line" x1="%f" y1="%f" x2="%f" y2="%f"/>' % (x+ms_width, y+net_offset_y, net_backbone_x, y+net_offset_y))
      out.append(translate(mlserver(mls,net_offset_y,leader_strver), x, y, "showGraphInfo('IM')"))
      net_end_y = y+net_offset_y
      y+=mls_height+mls_margin_top
    y-=mls_margin_top
  
  # Print chunk servers
  net_offset_y = 0.6 * cs_height
  odd = True
  cy_1st = cs_top
  cy_2nd = cs_top+cs_margin_top
  net_start_y = min(cy_1st,net_start_y)
  if len(chunkservers) > 0:
    # chunk servers tile
    x=net_backbone_x+net_backbone_hmargin
    y=cy_1st-tile_padding

    tile_width=cs_left_2nd+cs_width+tile_padding-(net_backbone_x+net_backbone_hmargin)
    tile_height=tile_padding+(math.ceil(len(chunkservers)/2)-1)*(cs_height+cs_margin_top)+cs_height+tile_padding+24
    if len(chunkservers) % 2 == 0:
      tile_height+=(cy_2nd-cy_1st)
    
    out.append('<rect x="%f" y="%f" width="%f" height="%f" rx="4" ry="4" class="tile-bckgrnd"/>' % (x, y, tile_width, tile_height))
    out.append(text('Chunkservers', x+8, y+tile_height-9, 'tile-title-left'))
    svg_h=max(svg_h, y+tile_height)
    svg_w=max(svg_w, x+tile_width)
    for cs in chunkservers:
      if odd:
        x = cs_left_1st
        y = cy_1st
        cy_1st += cs_height+cs_margin_top
      else:
        x = cs_left_2nd
        y = cy_2nd
        cy_2nd += cs_height+cs_margin_top
      net_end_y = max(net_end_y,y+net_offset_y)
      out.append('<line class="network-line" x1="%f" y1="%f" x2="%f" y2="%f"/>' % (x, y+net_offset_y, net_backbone_x, y+net_offset_y))
      out.append(translate(cserver(cs,net_offset_y,leader_strver), x, y, "showGraphInfo('CS-HD', 'CScsid=%s&HDdata=%s')" % (cs['id'],cs['id'])))
      odd = not odd
    cy_1st-=cs_margin_top
    cy_2nd-=cs_margin_top
  svg_h=max(svg_h, cy_1st)
  svg_h=max(svg_h, cy_2nd)
  
  net_start_y += net_offset_y
  # net_end_y += net_offset_y

  # Print network backbone line
  out.append('<line class="network-line" x1="%f" y1="%f" x2="%f" y2="%f"/>' % (net_backbone_x, net_start_y, net_backbone_x, net_end_y))

  # Adjust overall height, width and padding
  out.insert(0, '<g transform="matrix(1,0,0,1,%s,%s)">' % (padding, padding))
  svg_h+=2*padding
  svg_w+=2*padding
  out.insert(0, '<svg height="%d" width="%d">' % (svg_h, svg_w))

  out.append('</g>')
  out.append('</svg>')
  return "\n".join(out)



###################################################
### Prepare data for rendering and order render ###
###################################################

def render(dp, fields, vld):
  servers_warnings = []

  ci=dp.get_clusterinfo()
  
  # Prepare master servers info
  masterservers = []
  for ms in dp.get_masterservers(0): # sort by ip/port
    server = {
      "name": dp.cluster.masterhost,
      "ip": ms.strip,
      "port": ms.port,
      "strver": ms.strver,
      "statestr": ms.statestr,
      "cpuload": (ms.syscpu+ms.usercpu)/100.0,
      "memory": ms.memusage,
      "live": 1 if ms.is_active() else 0
    }
    if server['live']!=1 or ci.strver!=ms.strver or (not ms.statestr in ['LEADER', 'FOLLOWER', 'MASTER', '-']) or server['cpuload'] > thrsh_wrn:
      servers_warnings.append(ms.strip)
    masterservers.append(server)

  metaloggers = []
  for ml in dp.get_metaloggers(2): # sort by ip/port
    server = {
      "name": ml.host,
      "ip": ml.strip,
      "strver": ml.strver,
    }
    if ci.strver!=ml.strver:
      servers_warnings.append(ml.strip)
    metaloggers.append(server)
  
  chunkservers = []
  (hdds_all, scanhdds)=dp.get_hdds("ALL")
  for cs in dp.get_chunkservers(): # default sort by ip/port
    (hdds_status, hdds)=dp.cs_hdds_status(cs, hdds_all+scanhdds)
    if cs.is_maintenance_off() or not dp.cluster.leaderfound():
      maintenance = 0
    else:
      maintenance = 1
    srv = {
      "id": "%s:%s" % (cs.strip,cs.port),
      "name": cs.host,
      "strver": cs.strver,
      "ip": cs.strip,
      "port": cs.port,
      "live": 1 if cs.is_connected() else 0,     # if (cs.flags&1)!=0: <- dead server
      "maintenance": maintenance, # 1 - server in maintenance mode
      "queue": cs.queue,
      "queue_state": cs.queue_state,
      "queue_state_str": cs.queue_state_str,
      "queue_state_msg": cs.queue_state_msg,
      "flags": cs.flags,
      "hdd_reg_total": cs.total,
      "hdd_reg_used": cs.used,
      "hdd_rem_total": cs.tdtotal,
      "hdd_rem_used": cs.tdused,
      "hdds_status": hdds_status,
      "hdds": None #unused: hdds
      }
    chunkservers.append(srv)

    if srv['live']!=1 or srv['maintenance']!=0 or ci.strver!=cs.strver or srv['queue_state'] == CS_LOAD_OVERLOADED or (srv['hdd_reg_total']!=0 and srv['hdd_reg_used']/srv['hdd_reg_total'] > thrsh_wrn) or hdds_status!='ok':
      servers_warnings.append(cs.strip)
  
  # Prepare general cluster info 
  # Gather IOPS and W/R IOPS 
  sessions=dp.get_sessions(None) # don't sort
  rops_c_total = 0
  rops_l_total = 0
  wops_c_total = 0
  wops_l_total = 0
  ops_c_total = 0
  ops_l_total = 0
  for ses in sessions: #summing both last and current hour operations
    ops_c_total += sum(ses.stats_c)
    ops_l_total += sum(ses.stats_l)
    if (dp.stats_to_show>16):
      rops_c_total=ses.stats_c[16] # 17: read current hour
      wops_c_total=ses.stats_c[17] # 18: write current hour
      rops_l_total=ses.stats_l[16] # 17: read last hour
      wops_l_total=ses.stats_l[17] # 18: write last hour
  #if more than half of the minute passed (of current hour), calculate ops per second based on current hour stats
  if seconds_since_beginning_of_hour()>30.0: 
    ops = float(ops_c_total)/float(seconds_since_beginning_of_hour())
    rops = float(rops_c_total)/float(seconds_since_beginning_of_hour())
    wops = float(wops_c_total)/float(seconds_since_beginning_of_hour())
  else: # take last (previous) hour stats
    ops = float(ops_l_total)/3600.0
    rops = float(rops_l_total)/3600.0
    wops = float(wops_l_total)/3600.0
  if (dp.stats_to_show<=16):
    wops = None # this master doesn't support W/R stats
    ops = None

  # Gather throughput stats
  bread_ext = None
  bwrite_ext = None
  if dp.cluster.master().version_at_least(4,57,0): # mountbytrcvd, mountbytsent available from v.4.57.0
    # Get 60 minutes average of given (single row) master server chart data
    def get_mschart_60min_avg(no):
      raw = 0
      chrange = 0   # 1 minute ticks
      mccount = 60  # take 60 minutes (or less if 60 minutes not available)
      _,base,datadict = get_charts_multi_data(dp.cluster.master(), no*10+chrange, mccount)
      if datadict!=None and len(datadict)>0:
        val = 0
        count = 0
        ch1data,_,_,_,mul,div = datadict[0] # ch1data,ch2data,ch3data,ts,mul,div
        mul,div = adjust_muldiv(mul,div,base)
        chdata = charts_convert_data(ch1data,mul,div,raw)
        for chval in chdata:
          if chval!=None:
            val += chval
            count += 1
        if count>0:
          return val / count
        else:
          return None
        
    bread_ext = get_mschart_60min_avg(68) # 68 mountbytrcvd
    bwrite_ext = get_mschart_60min_avg(69) # 69 mountbytsent


  licence=dp.get_licence()
  licmaxsize = licence.licmaxsize if licence != None else None
  currentsize = licence.currentsize if licence != None else None
  # Get sums of missing, undergoal and endangered  (for all sclasses)
  (mx_summary,progressstatus)=dp.get_matrix_summary()
  clusterinfo = {
    "leaderfound" : 1 if dp.cluster.leaderfound() else 0,
    "strver": ci.strver,
    "totalspace": ci.totalspace,
    "availspace": ci.availspace,
    "freespace": ci.freespace,
    "dirs": ci.dirs,
    "files": ci.files,
    "trfiles": ci.trfiles,
    "trspace": ci.trspace,
    "chunks_progress": progressstatus,
    "chunks_total": mx_summary[MX_COL_TOTAL],
    "chunks_missing": mx_summary[MX_COL_MISSING],
    "chunks_undergoal": mx_summary[MX_COL_UNDERGOAL],
    "chunks_endangered": mx_summary[MX_COL_ENDANGERED],
    "servers_warnings": servers_warnings,
    "mounts": len(sessions),
    "ops": ops,	  #all (metadata and read and write) operations per second
    "rops": rops, #read operations per second
    "wops": wops, #write operations per second
    "bread_ext": bread_ext,   #traffic read by all CS - only from clients (ext)
    "bwrite_ext": bwrite_ext, #traffic write to all CS - only from clients (ext)
    "licmaxsize": licmaxsize, 
    "currentsize": currentsize
  }

  out = []
  out.append('<div id="mfsgraph-tile">')
  out.append(html_cluster_state(clusterinfo))
  out.append('<div id="mfsgraph">')
  out.append(svg_mfs_graph(masterservers, metaloggers, chunkservers, ci.strver))
  out.append('</div><!-- mfsgraph -->')
  out.append('<div id="mfsgraph-info"></div>')
  out.append('</div>')
  return out



from common.constants import *
from common.utilsgui import *

def render(dp, fields, vld):
	DEFAULT_MFLIMIT = 10
	MForder = fields.getint("MForder", 0)
	MFrev = fields.getint("MFrev", 0)
	MFlimit = fields.getint("MFlimit", DEFAULT_MFLIMIT)
	# if MFlimit!=0 and MFlimit<DEFAULT_MFLIMIT:
	# 	MFlimit = DEFAULT_MFLIMIT

	mchunks = dp.get_missing_chunks(MForder, MFrev)
	# mchunks = mchunks * 60
	# mchunks = mchunks[:5]
	mccnt = len(mchunks)
	if mccnt==0:
		return

	out = []
	# out.append("""<div class="tab_title">Missing files %s</div>""" % vld.issue('data_missing_files').icon())
	out.append("""<table class="acid_tab acid_tab_zebra_C1_C2 acid_tab_storageid_missingfiles">""")
	show_all = MFlimit==0 or MFlimit>=mccnt
	pagination = None

	if mccnt<DEFAULT_MFLIMIT:
		subtitle = ""
	else:
		if show_all:
			pagination = html_refresh_a('less', 'MFlimit', DEFAULT_MFLIMIT)
			subtitle = """- %u of %u entries - %s""" % (mccnt,mccnt,pagination)
		else:
			pagination = html_refresh_a('10 more', 'MFlimit', MFlimit+DEFAULT_MFLIMIT)
			if mccnt-MFlimit>100:
				pagination = "%s - %s" % (pagination, html_refresh_a('100 more', 'MFlimit', MFlimit+100))
			pagination = "%s - %s" % (pagination, html_refresh_a('all', 'MFlimit', '0'))
			if MFlimit>DEFAULT_MFLIMIT:
				pagination = "%s - %s" % (pagination, html_refresh_a('less', 'MFlimit', DEFAULT_MFLIMIT))
			subtitle = """- %u of %u entries - %s""" % (MFlimit,mccnt,pagination)

	out.append("""	<tr><th colspan="6">Missing files %s (gathered by the previous filesystem self-check loop) %s</th></tr>""" % (vld.issue('data_missing_files').icon(),subtitle))
	out.append("""	<tr>""")
	out.append("""		<th class="acid_tab_enumerate">#</th>""")
	out.append("""		<th>Paths</th>""")
	out.append("""		<th>Inode</th>""")
	out.append("""		<th>Index</th>""")
	out.append("""		<th>Chunk&nbsp;id</th>""")
	out.append("""		<th>Type&nbsp;of&nbsp;missing&nbsp;chunk</th>""")
	out.append("""	</tr>""")

	missingcount = 0
	for mc in mchunks:
		info = vld.get_missing_chunk_info(mc)
		if mccnt>0:
			if len(mc.paths)==0:
				if missingcount<MFlimit or MFlimit==0:
					out.append("""	<tr>""")
					out.append("""		<td align="right"></td>""")
					out.append("""		<td align="left"> * unknown path * (deleted file)</td>""")
					out.append("""		<td align="center">%u</td>""" % mc.inode)
					out.append("""		<td align="center">%u</td>""" % mc.indx)
					out.append("""		<td align="center" class="monospace">%s</td>""" % mc.get_chunkid_str())
					out.append("""		<td align="center">%s</td>""" % info.span_noicon(mc.get_mtype_str()))
					out.append("""	</tr>""")
				missingcount += 1
			else:
				for path in mc.paths:
					if missingcount<MFlimit or MFlimit==0:
						out.append("""	<tr>""")
						out.append("""		<td align="right"></td>""")
						out.append("""		<td align="left">%s</td>""" % path)
						out.append("""		<td align="center">%u</td>""" % mc.inode)
						out.append("""		<td align="center">%u</td>""" % mc.indx)
						out.append("""		<td align="center" class="monospace">%s</td>""" % mc.get_chunkid_str())
						out.append("""		<td align="center">%s</td>""" % info.span_noicon(mc.get_mtype_str()))
						out.append("""	</tr>""")
					missingcount += 1

	if pagination:
		txt = "all" if (MFlimit==0 or MFlimit>mccnt) else "%s of %s" % (MFlimit,mccnt)
		out.append("""<tfoot><tr><td colspan="6">Showing %s entries - %s</td></tr></tfoot>""" % (txt,pagination))
	out.append("""</table>""")
	return out


from common.constants import *
from common.utils import *
from common.utilsgui import *
import views.metaloggers

def render(dp, fields, vld):
	IMorder = fields.getint("IMorder", 0)
	IMrev = fields.getint("IMrev", 0)
	MBorder = fields.getint("MBorder", 0)
	MBrev = fields.getint("MBrev", 0)

	# update master servers delay times prior to getting the list of master servers
	highest_saved_metaversion, highest_metaversion_checksum = dp.cluster.update_masterservers_delays()

	mservers = dp.get_masterservers(IMorder, IMrev)
	mloggers = dp.get_metaloggers(MBorder, MBrev)

	out = []
	issues = vld.check_cluster_masters(mservers, len(dp.get_metaloggers()))
	if len(mloggers)>0 and len(mservers)>0:
		if len(mservers)==1:
			out.append("""<div class="tab_title">Metadata server (Master) and Metaloggers%s</div>""" % issues.span())
		else:
			out.append("""<div class="tab_title">Metadata servers (masters) and Metaloggers%s</div>""" % issues.span())
		out.append("""<table class="FR panel no-hover">""")
		out.append(""" <tr><td>""")		
	elif len(mservers)==1:
		out.append("""<div class="tab_title">Metadata server (Master)%s</div>""" % issues.span())
	elif len(mservers)>1:
		out.append("""<div class="tab_title">Metadata servers (masters)%s</div>""" % issues.span())
	elif len(mloggers)>0:
		out.append("""<div class="tab_title">Metadata loggers%s</div>""" % issues.span())
	else:
		out.append("""<div class="tab_title">Missing metadata servers%s</div>""" % issues.span())

	out.append("""<table class="acid_tab acid_tab_storageid_mfsmasters">""")
	out.append("""	<tr>""")
	out.append("""		<th rowspan="2" class="acid_tab_enumerate">#</th>""")
	out.append("""		<th rowspan="2">IP</th>""")
	out.append("""		<th rowspan="2">Version</th>""")
	out.append("""		<th rowspan="2">State</th>""")
	out.append("""		<th rowspan="2">Local time</th>""")
	out.append("""		<th colspan="3">Current metadata</th>""")
	out.append("""		<th colspan="2">Resources</th>""")
	out.append("""		<th colspan="5">Last metadata save</th>""")
	out.append("""		<th rowspan="2" data-tt="info_masters_exports_checksum">Exports<br/>checksum</th>""")
	out.append("""	</tr>""")
	out.append("""	<tr>""")
	out.append("""		<th>id</th>""")
	out.append("""		<th>version</th>""")
	out.append("""		<th>delay</th>""")
	out.append("""		<th>RAM</th>""")
	out.append("""		<th>CPU (sys&#8239;+&#8239;usr)</th>""")
	out.append("""		<th>time</th>""")
	out.append("""		<th>duration</th>""")
	out.append("""		<th>status</th>""")
	out.append("""		<th>version</th>""")
	out.append("""		<th>checksum</th>""")
	out.append("""	</tr>""")

	if len(mservers)==0:
		out.append("""	<tr><td colspan="10">Metadata (Master) servers not found! Check your DNS</td></tr>""")


	for ms in mservers:
		out.append("""	<tr>""")
		issues = vld.check_follower_version(ms)
		out.append("""		<td align="right"></td><td align="center" style="width:6%%;"><span class="sortkey">%s </span>%s</td><td align="center" style="width:6%%;"><span class="sortkey">%s</span>%s</td>""" % (ms.sortip,ms.strip,ms.sortver,issues.span(ms.strver.replace(" PRO", "<small> PRO</small>"))))
		issues = vld.check_ms_state(ms)
		if issues.any():
			out.append("""		<td align="center">%s</td>""" % issues.span(ms.statestr))
		else:
			out.append("""		<td align="center"><span class="STATECOLOR%u">%s</span></td>""" % (ms.statecolor, ms.statestr))
		if ms.usectime==None or ms.usectime==0:
			out.append("""		<td align="center">-</td>""")
		else:
			issues = vld.check_follower_clock(ms.secdelta)
			out.append("""		<td align="center">%s%s</td>""" % (datetime_to_str(ms.usectime//1000000), issues.span()))
		if ms.metaid!=None:
			issues = vld.check_follower_metaid(ms)
			out.append("""		<td align="center">%016X%s</td>""" % (ms.metaid, issues.span()))
		else:
			out.append("""		<td align="center">-</td>""")
		metaversion_str=decimal_number_html(ms.metaversion) if ms.metaversion>0 else '-'
		out.append("""		<td align="right">%s</td>""" % (metaversion_str))
		if ms.metadelay==None:
			out.append("""		<td align="right">-</td>""")
		else:
			issues = vld.check_follower_metadelay(ms.metadelay)
			out.append("""		<td align="right">%.0f&#8239;s%s</td>""" % (ms.metadelay, issues.span()))
		if ms.memusage>0:
			out.append("""		<td align="right"><span data-tt-info="%s&#8239;B">%s</span></td>""" % (decimal_number(ms.memusage),humanize_number(ms.memusage,"&nbsp;")))
		else:
			msg = vld.issue('ms_no_ram_info').span('n/a') if ms.is_active() else '-'
			out.append("""		<td align="center">%s</td>""" % msg)
		if ms.syscpu>0 or ms.usercpu>0:
			out.append("""		<td align="center"><span data-tt-info="%.3f%% (sys:%.3f%%, usr:%.3f%%)">%.1f%%&nbsp;(%.1f&#8239;+&#8239;%.1f)</span></td>""" % (ms.syscpu+ms.usercpu,ms.syscpu,ms.usercpu,ms.syscpu+ms.usercpu,ms.syscpu,ms.usercpu))
		else:
			msg = vld.issue('ms_no_cpu_info').span('n/a') if ms.is_active() else '-'
			out.append("""		<td align="center">%s</td>""" % msg)
		if ms.lastsuccessfulstore>0:
			issues = vld.check_ms_metadata_save_older(ms, highest_saved_metaversion)
			out.append("""		<td align="center">%s</td>""" % issues.span(datetime_to_str(ms.lastsuccessfulstore)))
			issues = vld.check_last_save_duration(ms)
			out.append("""		<td align="center"><span data-tt-info="%s">%s</span></td>""" % (timeduration_to_fullstr(ms.lastsaveseconds),issues.span(timeduration_to_shortstr(ms.lastsaveseconds,"&#8239;"))))
		else:
			out.append("""		<td align="center">-</td><td align="center">-</td>""")
		if ms.lastsuccessfulstore>0 or ms.lastsavestatus>0:
			issues = vld.check_ms_last_store_status(ms)
			txt='Background save' if ms.lastsavestatus==LASTSTORE_META_STORED_BG else "Got from other" if ms.lastsavestatus==LASTSTORE_DOWNLOADED else 'Foreground save' if ms.lastsavestatus==LASTSTORE_META_STORED_FG else "CRC background save" if ms.lastsavestatus==LASTSTORE_CRC_STORED_BG else "Unknown: %u" % ms.lastsavestatus
			out.append("""		<td align="center"><span class="%s" data-tt="%s">%s%s</span></td>""" % (issues.css_class(), issues.data_help(), txt, issues.icon()))
		else:
			out.append("""		<td align="center">-</td>""")
		if ms.lastsuccessfulstore>0 and (ms.lastsavestatus==LASTSTORE_META_STORED_BG or ms.lastsavestatus>=LASTSTORE_META_STORED_FG) and ms.lastsavemetaversion!=None:
			issues = vld.check_ms_metadata_save_older(ms, highest_saved_metaversion)
			out.append("""		<td align="right"><span class="%s" data-tt="%s">%s</span></td>""" % (issues.css_class(),issues.data_help(), decimal_number_html(ms.lastsavemetaversion)))
		else:
			out.append("""		<td align="center">-</td>""")
		if ms.lastsuccessfulstore>0 and (ms.lastsavestatus==LASTSTORE_META_STORED_BG or ms.lastsavestatus>=LASTSTORE_META_STORED_FG) and ms.lastsavemetaversion!=None and ms.lastsavemetachecksum!=None:
			issues = vld.check_ms_checksum_mismatch(ms, highest_saved_metaversion, highest_metaversion_checksum)
			out.append("""		<td align="center"><span class="%s" data-tt="%s">%08X %s</span></td>""" % (issues.css_class(),issues.data_help(),ms.lastsavemetachecksum,issues.icon()))
		else:
			out.append("""		<td align="center">-</td>""")
		if ms.exportschecksum!=None:
			issues = vld.check_follower_checksum(ms)
			out.append("""		<td align="center">%s</td>""" % issues.span(("%016X" % ms.exportschecksum)[:8]))
		else:
			out.append("""		<td align="center">-</td>""")
		out.append("""	</tr>""")	

	out.append("""</table>""")

	if len(mloggers)>0:
		out_mloggers = views.metaloggers.render(dp, fields, vld)
		out.extend(out_mloggers)

	if len(mloggers)>0 and len(mservers)>0:
		out.append("""</table>""") #panel table


	return out


from common.utils import *
from common.utilsgui import *
from common.models import *

def render(dp, fields, vld):
	OFsessionid = fields.getint("OFsessionid", 0)
	OForder = fields.getint("OForder", 0)
	OFrev = fields.getint("OFrev", 0)

	ALinode = fields.getint("ALinode", 0)
	ALorder = fields.getint("ALorder", 0)
	ALrev = fields.getint("ALrev", 0)

	sessionsdata = {}
	for ses in dp.get_sessions():
		if ses.sessionid>0 and ses.sessionid < 0x80000000:
			sessionsdata[ses.sessionid]=(ses.host,ses.info,ses.openfiles)
	out = []
	class_no_table=""
	if OFsessionid==0:
		class_no_table="no_table"
	out.append("""<form action="#"><div class="tab_title %s">Open files for client: """ % class_no_table)
	out.append("""<div class="select-fl"><select name="server" size="1" onchange="document.location.href='%s&OFsessionid='+this.options[this.selectedIndex].value">""" % fields.createrawlink({"OFsessionid":""}))
	if OFsessionid==0:
		out.append("""<option value="0" selected="selected"> select session</option>""")
	sessions = list(sessionsdata.keys())
	sessions.sort()
	for sessionid in sessions:
		host,info,openfiles = sessionsdata[sessionid]
		if OFsessionid==sessionid:
			out.append("""<option value="%s" selected="selected">%s: %s:%s (open files: ~%u)</option>""" % (sessionid,sessionid,host,info,openfiles))
		else:
			out.append("""<option value="%s">%s: %s:%s (open files: ~%u)</option>""" % (sessionid,sessionid,host,info,openfiles))
	out.append("""</select><span class="arrow"></span></div></div></form>""")

	if OFsessionid!=0:
		inodes = set()
		out.append("""<table class="acid_tab acid_tab_zebra_C1_C2 acid_tab_storageid_mfsopenfiles">""")
		out.append("""	<tr>""")
		out.append("""		<th class="acid_tab_enumerate">#</th>""")
		out.append("""		<th>Session&nbsp;id</th>""")
		out.append("""		<th>Host</th>""")
		out.append("""		<th>IP</th>""")
		out.append("""		<th>Mount&nbsp;point</th>""")
		out.append("""		<th>Inode</th>""")
		out.append("""		<th>Paths</th>""")
		out.append("""	</tr>""")
				
		for of in dp.get_openfiles(OFsessionid, OForder, OFrev):
			inodes.add(of.inode)
			for path in of.paths:
				out.append("""	<tr>""")
				out.append("""		<td align="right"></td>""")
				out.append("""		<td align="center">%u</td>""" % of.sessionid)
				out.append("""		<td align="left">%s</td>""" % of.host)
				out.append("""		<td align="center"><span class="sortkey">%s </span>%s</td>""" % (of.sortipnum,of.ipnum))
				out.append("""		<td align="left">%s</td>""" % htmlentities(of.info))
				out.append("""		<td align="center">%u</td>""" % of.inode)
				out.append("""		<td align="left">%s</td>""" % htmlentities(path))
				out.append("""	</tr>""")
		out.append("""</table>""")
	
		# Acquired locks for inode
		if ALinode not in inodes:
				ALinode = 0
		if len(inodes)>0:
			class_no_table = ""
			if ALinode == 0:
				class_no_table = "no_table"
			out.append("""<form action="#"><div class="tab_title %s">Acquired locks for inode: """ % class_no_table)
			out.append("""<div class="select-fl"><select name="server" size="1" onchange="document.location.href='%s&ALinode='+this.options[this.selectedIndex].value">""" % fields.createrawlink({"ALinode":""}))
			if ALinode==0:
				out.append("""<option value="0" selected="selected"> select inode</option>""")
			inodeslist = list(inodes)
			inodeslist.sort()
			for inode in inodeslist:
				if ALinode==inode:
					out.append("""<option value="%u" selected="selected">%u</option>""" % (inode,inode))
				else:
					out.append("""<option value="%u">%u</option>""" % (inode,inode))
			out.append("""</select><span class="arrow"></span></div></div></form>""")
			if ALinode!=0:
				out.append("""<table class="acid_tab acid_tab_zebra_C1_C2 acid_tab_storageid_acquiredlocks">""")
				out.append("""	<tr>""")
				out.append("""		<th class="acid_tab_enumerate">#</th>""")
				out.append("""		<th>Session&nbsp;id</th>""")
				out.append("""		<th>Host</th>""")
				out.append("""		<th>IP</th>""")
				out.append("""		<th>Mount&nbsp;point</th>""")
				out.append("""		<th>Lock type</th>""")
				out.append("""		<th>Owner id</th>""")
				out.append("""		<th>Pid</th>""")
				out.append("""		<th>Start</th>""")
				out.append("""		<th>End</th>""")
				out.append("""		<th>R/W</th>""")
				out.append("""	</tr>""")

				for al in dp.get_acquiredlocks(ALinode, ALorder, ALrev):
					out.append("""	<tr>""")
					out.append("""		<td align="right"></td>""")
					out.append("""		<td align="center">%u</td>""" % al.sessionid)
					out.append("""		<td align="left">%s</td>""" % al.host)
					out.append("""		<td align="center"><span class="sortkey">%s </span>%s</td>""" % (al.sortipnum,al.ipnum))
					out.append("""		<td align="left">%s</td>""" % htmlentities(al.info))
					out.append("""		<td align="center">%s</td>""" % al.locktype)
					out.append("""		<td align="right">%u</td>""" % al.owner)
					if al.pid==0 and al.start==0 and al.end==0:
						out.append("""		<td align="right">-1</td>""")
						out.append("""		<td align="right">0</td>""")
						out.append("""		<td align="right">EOF</td>""")
					else:
						out.append("""		<td align="right">%u</td>""" % al.pid)
						out.append("""		<td align="right">%u</td>""" % al.start)
						if al.end > 0x7FFFFFFFFFFFFFFF:
							out.append("""		<td align="right">EOF</td>""")
						else:
							out.append("""		<td align="right">%u</td>""" % al.end)
					out.append("""		<td align="center">%s</td>""" % ("Read (shared)" if al.ctype==MFS_LOCK_TYPE_SHARED else "Write (exclusive)" if al.ctype==MFS_LOCK_TYPE_EXCLUSIVE else "Unknown"))
					out.append("""	</tr>""")
				out.append("""</table>""")

	return out


import time
from common.constants import *
from common.utils import *
from common.utilsgui import *

def render(dp, fields, vld, readonly, selectable):
	HDperiod = fields.getint("HDperiod", 0)
	HDtime = fields.getint("HDtime", 0)
	HDorder = fields.getint("HDorder", 0)
	HDrev = fields.getint("HDrev", 0)
	HDdata = fields.getstr("HDdata", "")

	(hdds, scanhdds) = dp.get_hdds(HDdata, HDperiod, HDtime, HDorder, HDrev)
	out = []
	class_no_table = ""
	if len(hdds)==0 and len(scanhdds)==0:
		class_no_table = "no_table"
	if selectable:
		out.append("""<form action="#"><div class="tab_title %s">Disks, select: """ % class_no_table )
		out.append("""<div class="select-fl"><select name="server" size="1" onchange="document.location.href='%s&HDdata='+this.options[this.selectedIndex].value">""" % fields.createrawlink({"HDdata":""}))
		entrystr = []
		entrydesc = {}
		entrystr.append("ALL")
		entrydesc["ALL"] = "All disks"
		entrystr.append("ERR")
		entrydesc["ERR"] = "Disks with errors only"
		entrystr.append("NOK")
		entrydesc["NOK"] = "Disks with status other than ok"
		servers=dp.get_chunkservers()
		for cs in servers:
			hostx = resolve(cs.strip,0,UNRESOLVED)
			if hostx==UNRESOLVED:
				host = ""
			else:
				host = " / "+hostx
			entrystr.append(cs.hostkey)
			entrydesc[cs.hostkey] = "Server: %s%s" % (cs.hostkey,host)
		if HDdata not in entrystr:
			out.append("""<option value="" selected="selected">Pick option...</option>""")
		for estr in entrystr:
			if estr==HDdata:
				out.append("""<option value="%s" selected="selected">%s</option>""" % (estr,entrydesc[estr]))
			else:
				out.append("""<option value="%s">%s</option>""" % (estr,entrydesc[estr]))
		out.append("""</select><span class="arrow"></span></div>""")
		out.append("""</div></form>""")
	else:
		if len(hdds)>0 or len(scanhdds)>0:
			out.append("""<div class="tab_title %s">Disks</div>""")

	if len(hdds)==0 and len(scanhdds)==0:
		return out
	
	# prepare headers
	out.append("""<table class="acid_tab acid_tab_zebra_C1_C2 acid_tab_storageid_mfshdd" id="mfshdd">""")
	out.append("""	<tr>""")
	out.append("""		<th colspan="5" rowspan="2" class="knob-cell">""")

	options=[(-50, "IP address", None, "acid_tab.switchdisplay('mfshdd','hddaddrname_vis',0);"),
			(-130,"Server name", None,"acid_tab.switchdisplay('mfshdd','hddaddrname_vis',1);")]
	out.append(html_knob_selector("hddaddrname_vis",12,(130,41),(110,20),options))

	options=[(-50,"Last minute", None, "acid_tab.switchdisplay('mfshdd','hddperiod_vis',0);"),
			(50,"Last hour", None, "javascript:acid_tab.switchdisplay('mfshdd','hddperiod_vis',1);"),
			(130,"Last day", None, "javascript:acid_tab.switchdisplay('mfshdd','hddperiod_vis',2);")]
	out.append(html_knob_selector("hddperiod_vis",12,(210,41),(110,20),options))

	options=[(50, "Max time", None,"javascript:acid_tab.switchdisplay('mfshdd','hddtime_vis',0);"),
			(130,"Avg time", None,"javascript:acid_tab.switchdisplay('mfshdd','hddtime_vis',1)")]
	out.append(html_knob_selector("hddtime_vis",12,(110,41),(20,20),options))

	out.append("""		</th>""")
	out.append("""		<th colspan="8">""")
	out.append("""			<span class="hddperiod_vis0">I/O stats - last minute</span>""")
	out.append("""			<span class="hddperiod_vis1">I/O stats - last hour</span>""")
	out.append("""			<span class="hddperiod_vis2">I/O stats - last day</span>""")
	out.append("""		</th>""")
	out.append("""		<th colspan="3" rowspan="2">Capacity</th>""")
	out.append("""	</tr>""")
	out.append("""	<tr>""")
	out.append("""		<th colspan="2"><span data-tt-help="Average data transfer speed">transfer</span></th>""")
	out.append("""		<th colspan="3">""")
	out.append("""			<span class="hddtime_vis0" data-tt-help="Maximum time of read or write one chunk block (up to 64kB)">max operation time</span>""")
	out.append("""			<span class="hddtime_vis1" data-tt-help="Average time of read or write one chunk block (up to 64kB)">average operation time</span>""")
	out.append("""		</th>""")
	out.append("""		<th colspan="3"><span data-tt-help="Number of chunk block operations: reads, writes and fsyncs respectively">number of ops</span></th>""")
	out.append("""	</tr>""")
	out.append("""	<tr>""")
	out.append("""		<th class="acid_tab_enumerate">#</th>""")
	out.append("""		<th class="acid_tab_level_1"><span class="hddaddrname_vis0">IP</span><span class="hddaddrname_vis1">server name</span> and path</th>""")
	out.append("""		<th>chunks</th>""")
	out.append("""		<th>last error</th>""")
	out.append("""		<th>status</th>""")
	out.append("""		<th class="acid_tab_level_1">read</th>""")
	out.append("""		<th class="acid_tab_level_1">write</th>""")
	out.append("""		<th class="acid_tab_level_2">read</th>""")
	out.append("""		<th class="acid_tab_level_2">write</th>""")
	out.append("""		<th class="acid_tab_level_2">fsync</th>""")
	out.append("""		<th class="acid_tab_level_1">read</th>""")
	out.append("""		<th class="acid_tab_level_1">write</th>""")
	out.append("""		<th class="acid_tab_level_1">fsync</th>""")
	out.append("""		<th>used</th>""")
	out.append("""		<th>total</th>""")
	out.append("""		<th class="progbar">% used</th>""")
	out.append("""	</tr>""")

	# calculate some stats
	usedsum = {}
	totalsum = {}
	hostavg = {}
	for _,hdd in hdds+scanhdds:
		if hdd.hostkey not in usedsum:
			usedsum[hdd.hostkey]=0
			totalsum[hdd.hostkey]=0
			hostavg[hdd.hostkey]=0
		if hdd.flags&CS_HDD_SCANNING==0 and hdd.total and hdd.total>0:
			usedsum[hdd.hostkey]+=hdd.used
			totalsum[hdd.hostkey]+=hdd.total
			if totalsum[hdd.hostkey]>0:
				hostavg[hdd.hostkey] = (usedsum[hdd.hostkey] * 100.0) / totalsum[hdd.hostkey]

	for sf,hdd in hdds+scanhdds:
		if not hdd.is_valid():
			lerror= '-'
		elif not hdd.has_errors():
			lerror = 'no errors'
		else:
			errtimetuple = time.localtime(hdd.errtime)
			errtimelong = time.strftime("%Y-%m-%d %H:%M:%S",errtimetuple)
			errtimeshort = time.strftime("%Y-%m-%d %H:%M",errtimetuple)
			if readonly:
				lerror = '<span data-tt-warning="Error at %s <br/>on chunk #%016X">%s</span>' % (errtimelong,hdd.errchunkid,errtimeshort)
			else:
				if hdd.flags&CS_HDD_DAMAGED:
					lerror = """<span data-tt-warning="Error occurred at %s<br/>on chunk #%016X<br/>You can't dismiss it because this disk<br/>is marked as damaged.">%s</span>""" % (errtimelong,hdd.errchunkid,errtimeshort)
				else:
					found_cs = list(filter(lambda cs: cs.hostkey == hdd.hostkey, dp.get_chunkservers(None)))
					cs = found_cs[0] if len(found_cs)==1 else None
					if cs and cs.version>=(4,33,0):
						lerror = """<a href="%s" data-tt-warning="Click, to dismiss error that occurred<br/>at %s <br/>on chunk #%016X">%s</a>""" % (fields.createhtmllink({"CSclearerrors":hdd.clearerrorarg}),errtimelong,hdd.errchunkid,errtimeshort)
					else:
						lerror = """<span data-tt-warning="Error occurred at %s<br/>on chunk #%016X<br/>You can't dismiss it because this chunkserver<br/>is too old (min. v.4.33 is required).<br/>It is recommended to upgrade chunkserver.">%s</span>""" % (errtimelong,hdd.errchunkid,errtimeshort)

		if hdd.is_valid():
			chunkscnttxt=decimal_number_html(hdd.chunkscnt)
			usedtxt = humanize_number(hdd.used,"&nbsp;")
			totaltxt = humanize_number(hdd.total,"&nbsp;")
		else:
			chunkscnttxt = '-'
			usedtxt = '-'
			totaltxt = '-'
		out.append("""	<tr>""")
		out.append("""		<td align="right"></td>""")
		out.append("""		<td align="left"><span class="hddaddrname_vis0"><span class="sortkey">%s</span>%s</span><span class="hddaddrname_vis1">%s</span></td>""" % (htmlentities(hdd.sortippath),htmlentities(hdd.ippath).replace(":/", ": /"),htmlentities(hdd.hostpath).replace(":/", ": /")))
		issues = vld.check_hdd_status(hdd)
		clz = "OK" if issues.len()==0 else "MFRREADY" if hdd.mfrstatus==MFRSTATUS_READY else ""
		out.append("""		<td align="right">%s</td><td><span class="sortkey">%u</span>%s</td><td>%s</td>""" % (chunkscnttxt,hdd.errtime,lerror,issues.span(hdd.get_status_str(),clz)))
		validdata = [1,1,1]
		for i in range(3):
			if hdd.rbw[i]==0 and hdd.wbw[i]==0 and hdd.usecreadmax[i]==0 and hdd.usecwritemax[i]==0 and hdd.usecfsyncmax[i]==0 and hdd.rops[i]==0 and hdd.wops[i]==0:
				validdata[i] = 0
		# rbw
		out.append("""		<td align="right">""")
		for i in range(3):
			out.append("""			<span class="hddperiod_vis%u">""" % i)
			if validdata[i]:
				out.append("""				<span class="sortkey">%u </span><span data-tt-info="%s B/s">%s/s</span>""" % (hdd.rbw[i],decimal_number(hdd.rbw[i]),humanize_number(hdd.rbw[i],"&nbsp;")))
			else:
				out.append("""				<span><span class="sortkey">-1</span>&nbsp;</span>""")
			out.append("""			</span>""")
		out.append("""		</td>""")
		# wbw
		out.append("""		<td align="right">""")
		for i in range(3):
			out.append("""			<span class="hddperiod_vis%u">""" % i)
			if validdata[i]:
				out.append("""				<span class="sortkey">%u </span><span data-tt-info="%s B/s">%s/s</span>""" % (hdd.wbw[i],decimal_number(hdd.wbw[i]),humanize_number(hdd.wbw[i],"&nbsp;")))
			else:
				out.append("""				<span><span class="sortkey">-1</span>&nbsp;</span>""")
			out.append("""			</span>""")
		out.append("""		</td>""")
		# readtime
		out.append("""		<td align="right">""")
		for i in range(3):
			out.append("""			<span class="hddperiod_vis%u">""" % i)
			if validdata[i]:
				out.append("""				<span class="hddtime_vis0">%s us</span>""" % decimal_number_html(hdd.usecreadmax[i]))
				out.append("""				<span class="hddtime_vis1">%s us</span>""" % decimal_number_html(hdd.usecreadavg[i]))
			else:
				out.append("""				<span><span class="sortkey">-1 </span>&nbsp;</span>""")
			out.append("""			</span>""")
		out.append("""		</td>""")
		# writetime
		out.append("""		<td align="right">""")
		for i in range(3):
			out.append("""			<span class="hddperiod_vis%u">""" % i)
			if validdata[i]:
				out.append("""				<span class="hddtime_vis0">%s us</span>""" % decimal_number_html(hdd.usecwritemax[i]))
				out.append("""				<span class="hddtime_vis1">%s us</span>""" % decimal_number_html(hdd.usecwriteavg[i]))
			else:
				out.append("""				<span><span class="sortkey">-1</span>&nbsp;</span>""")
			out.append("""			</span>""")
		out.append("""		</td>""")
		# fsynctime
		out.append("""		<td align="right">""")
		for i in range(3):
			out.append("""			<span class="hddperiod_vis%u">""" % i)
			if validdata[i]:
				out.append("""				<span class="hddtime_vis0">%s us</span>""" % decimal_number_html(hdd.usecfsyncmax[i]))
				out.append("""				<span class="hddtime_vis1">%s us</span>""" % decimal_number_html(hdd.usecfsyncavg[i]))
			else:
				out.append("""				<span><span class="sortkey">-1</span>&nbsp;</span>""")
			out.append("""			</span>""")
		out.append("""		</td>""")
		# rops
		out.append("""		<td align="right">""")
		for i in range(3):
			out.append("""			<span class="hddperiod_vis%u">""" % i)
			if validdata[i]:
				if hdd.rops[i]>0:
					bsize = hdd.rbytes[i]/hdd.rops[i]
				else:
					bsize = 0
				out.append("""				<span data-tt-info="average block size: %u B">%s</span>""" % (bsize,decimal_number_html(hdd.rops[i])))
			else:
				out.append("""				<span><span class="sortkey">-1</span>&nbsp;</span>""")
			out.append("""			</span>""")
		out.append("""		</td>""")
		# wops
		out.append("""		<td align="right">""")
		for i in range(3):
			out.append("""			<span class="hddperiod_vis%u">""" % i)
			if validdata[i]:
				if hdd.wops[i]>0:
					bsize = hdd.wbytes[i]/hdd.wops[i]
				else:
					bsize = 0
				out.append("""				<span data-tt-info="average block size: %u B">%s</span>""" % (bsize,decimal_number_html(hdd.wops[i])))
			else:
				out.append("""				<span><span class="sortkey">-1</span>&nbsp;</span>""")
			out.append("""			</span>""")
		out.append("""		</td>""")
		# fsyncops
		out.append("""		<td align="right">""")
		for i in range(3):
			out.append("""			<span class="hddperiod_vis%u">""" % i)
			if validdata[i]:
				out.append("""				%s""" % decimal_number_html(hdd.fsyncops[i]))
			else:
				out.append("""				<span><span class="sortkey">-1</span>&nbsp;</span>""")
			out.append("""			</span>""")
		out.append("""		</td>""")
		if hdd.flags&CS_HDD_SCANNING:
			out.append("""		<td colspan="3" align="right"><span class="sortkey">0 </span><div class="PROGBOX"><div class="PROGCOVER" style="width:%.0f%%;"></div><div class="PROGVALUE"><span>%.0f%% scanned</span></div></div></td>""" % (100.0-hdd.used,hdd.used))
		else:
			issues = vld.check_hdd_usage(hdd)
			out.append("""		<td align="right"><span class="sortkey">%u</span><span data-tt-info="%s B">%s</span></td><td align="right"><span class="sortkey">%u </span><span data-tt-info="%s B">%s</spana></td>""" % (hdd.used,decimal_number(hdd.used),issues.span(usedtxt),hdd.total,decimal_number(hdd.total),totaltxt))
			if hdd.total>0:
				usedpercent = (hdd.used*100.0)/hdd.total
				avgpercent = hostavg[hdd.hostkey]
				if totalsum[hdd.hostkey]!=hdd.total: 
					if avgpercent-usedpercent>=0.1:
						diffstr = "%.1f pp. lower than '%s' chunkserver's average (%.1f%%)" % (avgpercent-usedpercent, hdd.hoststr, avgpercent)
					elif usedpercent-avgpercent>=0.1:
						diffstr = "%.1f pp. higher than '%s' chunkserver's average (%.1f%%)" % (usedpercent-avgpercent, hdd.hoststr, avgpercent)
					else:
						diffstr = "equal to '%s' chunkserver's average" % hdd.hoststr
				else:
					diffstr = ''
				out.append("""		<td align="center"><span class="sortkey">%.10f </span><div class="PROGBOX"><div class="PROGCOVER" style="width:%.2f%%;"></div><div class="PROGAVG" style="width:%.2f%%"></div><div class="PROGVALUE"><span><span data-tt-info="%s">%.1f</span></span></div></div></td>""" % (usedpercent,100.0-usedpercent,avgpercent,diffstr,usedpercent))
			else: #skip for single disk chunkserver
				out.append("""		<td align="center"><span class="sortkey">-1 </span><div class="PROGBOX"><div class="PROGCOVER" style="width:100%;"></div><div class="PROGVALUE"><span>-</span></div></div></td>""")
		out.append("""	</tr>""")				
	out.append("""</table>""")

	return out



import subprocess
import os
import json
import time
import traceback
import re

from common.constants import *

def print_render(fields, masterhost, masterport):
	print("Content-Type: text/plain; charset=UTF-8\r\n\r")
	print('# HELP mfs_cgi_info MooseFS CGI general info')
	print('# TYPE mfs_cgi_info gauge')
	print("""mfs_cgi_info{version="%s"} 1.0""" % VERSION)
	def filter_letters_and_commas(input_string):
		# Use list comprehension to include only uppercase letters and commas
		filtered_characters = [char for char in input_string if (char.isalpha() or char == '_' or char == ',')]
		# Join the filtered characters into a new string
		return ''.join(filtered_characters)
	scope = fields.getvalue("scope", default="default")
	if scope is not None:
		if scope.lower() == 'none':
			scope = None 
		else:
			scope = filter_letters_and_commas(scope.upper()).split(",")
	mastercharts = fields.getvalue("mastercharts", default="all")
	if mastercharts is not None:
		if mastercharts.lower() == 'none':
			mastercharts = None
		else:
			mastercharts = filter_letters_and_commas(mastercharts.lower()).split(",")
	else:
		mastercharts = ['all']
	cscharts = fields.getvalue("cscharts", default="all")
	if cscharts is not None:
		if cscharts.lower() == 'none':
			cscharts = None
		else:
			cscharts = filter_letters_and_commas(cscharts.lower()).split(",")
	else:
		cscharts = ['all']
	prefix_whitelist = fields.getvalue("prefix_whitelist", default="")
	if prefix_whitelist is not None and prefix_whitelist:
		prefix_whitelist = filter_letters_and_commas(prefix_whitelist.lower()).split(",")
	else:
		prefix_whitelist = None
	prefix_blacklist = fields.getvalue("prefix_blacklist", default="")
	if prefix_blacklist is not None and prefix_blacklist:
		prefix_blacklist = filter_letters_and_commas(prefix_blacklist.lower()).split(",")
	else:
		prefix_blacklist = None
	
	print(MFSMetrics.get_metrics(masterhost=masterhost, masterport=masterport, scope=scope, mastercharts=mastercharts, cscharts=cscharts, prefix_whitelist=prefix_whitelist, prefix_blacklist=prefix_blacklist))


class MFSMetrics:
	def __init__(self, prefix_whitelist, prefix_blacklist):
		self.prefix_whitelist = prefix_whitelist
		self.prefix_blacklist = prefix_blacklist

		# exclude some keys from metrics: all chunks table doesn't fit well into prometheus metrics, so we exclude it and export summary only
		default_blacklist = [
					# 'mfs_info_missing',
					'mfs_info_chunks_allchunks','mfs_info_chunks_regularchunks',
					'mfs_info_chunks_allchunks_copies','mfs_info_chunks_regularchunks_copies',
					'mfs_info_chunks_allchunks_ec8','mfs_info_chunks_regularchunks_ec8',
					'mfs_info_chunks_allchunks_ec4','mfs_info_chunks_regularchunks_ec4']
		if self.prefix_blacklist is None:
			self.prefix_blacklist = default_blacklist
		else:
			self.prefix_blacklist += default_blacklist

		self.help_dict = []
		with open(os.path.join(os.path.dirname(os.path.abspath(__file__)),'../assets/help.json'), 'r', encoding='utf-8') as file:
			self.help_dict = json.load(file)

	# Calls 'mfscli' with appropriate (scope, mastercharts, cscharts) parameters and generates Prometheus metrics restricted by prefix_whitelist and prefix_blacklist
	@classmethod
	def get_metrics(clz, masterhost, masterport, scope, mastercharts, cscharts, prefix_whitelist, prefix_blacklist):
		generator = MFSMetrics(prefix_whitelist, prefix_blacklist)
		return generator.generate_metrics(masterhost, masterport, scope, mastercharts, cscharts)

	# 	-SIN : show full master info 			   - NOT allowed because includes possibly too long list of missing chunks/files (SMF)
	# 	-SIM : show only masters states
	# 	-SLI : show only licence info
	# 	-SIG : show only general master (leader) info
	# 	-SMU : show only master memory usage
	# 	-SIC : show only chunks info (target/current redundancy level matrices)
	# 	-SIL : show only loop info (with messages)
	# 	-SMF : show only missing chunks/files - NOT allowed because possibly too long list of missing chunks/files
	# 	-SCS : show connected chunk servers
	# 	-SMB : show connected metadata backup servers
	# 	-SHD : show hdd data
	# 	-SEX : show exports                 - NOT allowed because not much interesting numeric data here
	# 	-SMS : show active mounts           - NOT allowed because not much interesting numeric data here
	# 	-SRS : show resources               - NOT allowed because possibly too long list of open files and locks
	# 	-SSC : show storage classes
	# 	-SPA : show patterns override data  - NOT allowed because not much interesting numeric data here
	# 	-SOF : show only open files         - NOT allowed because possibly too long list of open files
	# 	-SAL : show only acquired locks     - NOT allowed because possibly too long list of acquired locks
	# 	-SMO : show operation counters
	# 	-SQU : show quota info
	# 	-SMC : show master charts data      - NOT allowed as scope because covered with cscharts
	# 	-SCC : show chunkserver charts data - NOT allowed as scope because covered with mastercharts

	def generate_metrics(self, masterhost, masterport, scope, mastercharts, cscharts):
		try:
			command = ["mfscli", "-j"]
			allowed_scopes = ["SIM", "SLI", "SIG", "SMU", "SIC", "SIL", "SCS", "SMB", "SHD", "SSC", "SMO", "SQU"]
			default_scopes = ["SIM", "SLI", "SIG",        "SIC", "SIL", "SCS", "SMB", "SHD"                     ]
			show_anything = False
			if scope == ["DEFAULT"]:
				for s in default_scopes:
					command.append("-"+str(s))
					show_anything = True
			elif scope is not None:
				for s in scope:
					if s in allowed_scopes:
						command.append("-"+str(s))
						show_anything = True
			if mastercharts is not None:
				command.append("-SMC")
				command.append("-a")
				command.append("1")
				command.append("-b")
				command.append(",".join(mastercharts))
				show_anything = True
			if cscharts is not None:
				command.append("-SCC")
				command.append("-c")
				command.append("1")
				command.append("-d")
				command.append(",".join(cscharts))
				show_anything = True
			if masterhost is not None:
				command.append("-H")
				command.append(str(masterhost))
			if masterport is not None:
				command.append("-P")
				command.append(str(masterport))
			# print("Command: ", command)
			# if len(command) == 2:
			# 	return "Error: No valid scope selected for MooseFS metrics (see mfscli '-S__' argument documentation), actual: "+str(scope)+" allowed: "+str(allowed_scopes)
			# return f"{scope} {mastercharts} {cscharts} {masterhost} {masterport}"
			if not show_anything:
				return ""
			start_time = time.time()
			proc = subprocess.Popen(command, env=self.make_env(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
			# proc = subprocess.Popen(["mfscli", "-j", "-SIN", "-SCS", "-SHD", "-SQU", "-SMC", "-SMO", "-SMS", "-a", "1", "-b", "all", "-SCC", "-c", "1", "-d", "all"],env=clz.make_env(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
			json_string = ""
			for line in proc.stdout:
				json_string+=line.decode('utf-8')
			execution_time_seconds = time.time() - start_time
			print('# HELP mfs_cli_execution_time CLI (mfscli) command execution time [seconds] used to fetch metrics')
			print('# TYPE mfs_cli_execution_time gauge')
			print("""mfs_cli_execution_time{command="%s"} %f""" % (" ".join(command),execution_time_seconds))
			# output += "Errors:\n"
			# for line in proc.stderr:
			# 	output+=line.decode('ascii')
			proc.stdout.close()
			proc.stderr.close()
			json_data = json.loads(json_string)
			return self.json2metrics(json_data)
		except json.JSONDecodeError as e:
			print("Failed to parse JSON: "+str(e))
			print("First line of JSON: "+json_string.split('\n')[0])
			print("CLI command for retrieving data: "+" ".join(command))
		except Exception as e:
			return "Error: "+str(e)+"\nCLI command: "+" ".join(command)+"\n"+traceback.format_exc()
	
	def make_env(self):
		env = {}
		env['PATH'] = BIN_PATH+":"+os.environ['PATH']
		return env
	
	def json2metrics(self, json_data):
		prefix = 'mfs'
		out = self.iterable2metrics(prefix,json_data['dataset'])
		return out

	# Recursively converts a json subset to prometheus metrics
	def iterable2metrics(self, prefix, subset, ids=None):
		if prefix == 'mfs_mastercharts' or prefix == 'mfs_cscharts':
			return self.charts2metrics(prefix, subset, ids)

		out = ""
		if isinstance(subset, dict):
			out += self.strings2metrics(prefix, subset, ids)
			for key in subset:
				out += self.iterable2metrics(prefix+'_'+key.lower(), subset[key], ids)
		elif isinstance(subset, list):
			for index,element in enumerate(subset):
				ids_this = self.prepare_ids(ids, index, element)
				out += self.iterable2metrics(prefix, element, ids_this)
		elif isinstance(subset, (int, float, bool)):
			out += self.number2metrics(prefix, subset, ids)
		return out
	
	# Converts a list of mastercharts to prometheus metrics
	def charts2metrics(self, prefix, subset, ids):
		out = ''
		for index,element in enumerate(subset):
			# check if we have any valid data?
			if element['data_ranges'] is not None and '0' in element['data_ranges'] and 'data_array_1' in element['data_ranges']['0']:
				value = element['data_ranges']['0']['data_array_1'][0]
				if (value == None):
					continue
				ids_this = self.prepare_ids(ids, index, element,['master','chunkserver'])	
				key = (prefix+'_'+element['name']).lower()
				if self.is_listed(key):
					out += self.help_string(key, element['description'])
					out += '# TYPE ' + key + ' gauge\n'
					out += key +'{'+self.ids2metrics(ids_this)+'}' + ' ' + str(value)+'\n'
		return out

	# Converts a number to a prometheus gauge metric
	def number2metrics(self, key, number, ids):
		if not self.is_listed(key):
			return ''
		out = ''
		out += self.help_string(key)
		out += '# TYPE ' + key + ' gauge\n'+key
		if ids is not None:
			out += '{'+self.ids2metrics(ids)+'}'
		if isinstance(number, bool):
			number = int(number)
		return out+' '+str(number)+'\n'
	
	# Prepares a list of labels for a given element - they serve as a unique identifier for a metric
	def prepare_ids(self, ids, index, element, identifiers=None):
		if identifiers is None:
			identifiers=['ip', 'port', 'hostname', 'name', 'id', 'csid', 'path', 'hostname_path', 'sclassname', 'state'] #, 'ip_path'
		if ids is None:
			ids = {}
		# if ids is None: don't use 'no' 
		# 	ids = {'no': str(index)}
		# else:
		# 	ids['no'] = str(ids['no'])+'_'+str(index)
		if isinstance(element, dict):
			for identifier in identifiers:
				if identifier in element:
					ids[identifier]=element[identifier]
		return ids

	# Converts a list of identifiers (for a lists of values) to a prometheus labels
	def ids2metrics(self, ids):
		out = ''
		for id_name in ids:
			if out != '':
				out+=','
			out+= self.safe_string(id_name)+'="' + str(ids[id_name]).replace('"',"'")+'"' 
		return out

	def safe_string(self, string):
		# replace all non-alphanumeric characters with _ and make sure it doesn't starts with a number
		if not string[0].isalpha():
			string = '_'+string
		return ''.join(e for e in string if e.isalnum() or e == '_')
	
	# Prepares a list of all strings in a subset as a prometheus label
	def strings2metrics(self, prefix, subset, ids):
		prefix += '_misc'
		if not self.is_listed(prefix):
			return ''
		strings = ''
		for key in subset:
			if ids is not None and key in ids.keys():
				continue # skip string if it is already in identifier
			if isinstance(subset[key], (str)) and not key.endswith('_human'):
				if strings !='':
					strings+=','
				strings += self.safe_string(key)+'="'+subset[key].replace('"',"'")+'"'

		if strings != '':
			labels = '{'
			if ids is not None:
				labels += self.ids2metrics(ids)+','
			labels += strings+'}'
			return self.help_string(prefix)+'# TYPE ' + prefix + ' gauge\n'+prefix+labels+' 1\n'
		return ''
	
	def whitelisted(self, prefix):
		# whitelist is present, so we only include metrics that start with the whitelisted prefixes
		if self.prefix_whitelist is None:
			return True
		for prefix_w in self.prefix_whitelist:
			if prefix.startswith(prefix_w):
				return True
		return False

	def blacklisted(self, prefix):
		# blacklist is present, so we exclude metrics that start with the blacklisted prefixes
		if self.prefix_blacklist is None:
			return False
		for prefix_b in self.prefix_blacklist:
			if prefix.startswith(prefix_b):
				return True
		return False
	
	# Filters out metrics based on the prefix whitelist and blacklist
	def is_listed(self, prefix):
		if not self.whitelisted(prefix):
			return False
		if self.blacklisted(prefix):
			return False
		return True

	def help_string(self, key, default=None):
		# skip first 4 letters as help_dict doesn't use 'mfs_' prefix
		if key[4:] in self.help_dict:
			help_msg = ''
			if 'title' in self.help_dict[key[4:]]: 
				help_msg += self.help_dict[key[4:]]['title'] 
				if help_msg != '' and help_msg[-1] == '.': # remove dot if title ends with it
					help_msg = help_msg[:-1]

			if 'unit' in self.help_dict[key[4:]]:
				help_msg += ' ['+self.help_dict[key[4:]]['unit']+']'
			# if 'description' in self.help_dict[key[4:]]:
			# 	if help_msg != '' and help_msg[-1] != '.': # add a dot if title doesn't end with it
			# 		help_msg += '. '
			# 	help_msg += self.help_dict[key[4:]]['description']
			# remove any html tags, leaving only text between them
			help_msg = re.sub(r'<[^>]*>', '', help_msg)
			return '# HELP '+ key + ' ' + help_msg + '\n'
		elif default is not None:
			return '# HELP '+ key + ' ' + default + '\n'
		else:
			return ''



from common.constants import *
from common.utils import *
from common.utilsgui import *

def render(dp, fields, vld):
	hsc = dp.get_health_selfcheck()
	mchunks = dp.get_missing_chunks(None)
	show_mfiles = len(mchunks)>0 
	out = []
	issues = vld.check_health_selfcheck(hsc)
	out.append("""<div class="tab_title">Filesystem self-check %s </div>""" % issues.icon())
	if show_mfiles:
		out.append("""<table class="FR panel no-hover">""")
		out.append(""" <tr><td>""")
	out.append("""<table class="FR no-hover">""")
	
	out.append("""	<tr>""")
	out.append("""		<th colspan="2">Self-check loop</th>""")
	out.append("""		<th colspan="5">Files (health)</th>""")
	out.append("""		<th colspan="3">Chunks (health)</th>""")
	out.append("""	</tr>""")

	out.append("""	<tr>""")
	out.append("""		<th style="min-width: 100px;">start time</th>""")
	out.append("""		<th style="min-width: 100px;">end time</th>""")
	out.append("""		<th style="min-width: 100px;">checked</th>""")
	out.append("""		<th style="min-width: 100px;">missing</th>""")
	out.append("""		<th style="min-width: 100px;">undergoal</th>""")
	out.append("""		<th style="min-width: 100px;">missing in trash</th>""")
	out.append("""		<th style="min-width: 100px;">missing sustained</th>""")
	out.append("""		<th style="min-width: 100px;">checked</th>""")
	out.append("""		<th style="min-width: 100px;">missing</th>""")
	out.append("""		<th style="min-width: 100px;">undergoal</th>""")
	out.append("""	</tr>""")
	if hsc.loopstart>0:
		out.append("""	<tr>""")
		out.append("""		<td align="center">%s</td>""" % (datetime_to_str(hsc.loopstart),))
		out.append("""		<td align="center">%s</td>""" % (datetime_to_str(hsc.loopend),))
		out.append("""		<td align="center">%s</td>""" % decimal_number_html(hsc.files))
		out.append("""		<td align="center"><span class="%s">%s</span></td>""" % ("" if hsc.mfiles==0 else "MISSING", decimal_number_html(hsc.mfiles)))
		out.append("""		<td align="center"><span class="%s">%s</span></td>""" % ("" if hsc.ugfiles==0 else "UNDERGOAL", decimal_number_html(hsc.ugfiles)))
		out.append("""		<td align="center">%s</td>""" % decimal_number_na_html(hsc.mtfiles))
		out.append("""		<td align="center">%s</td>""" % decimal_number_na_html(hsc.msfiles))
		out.append("""		<td align="center">%s</td>""" % decimal_number_html(hsc.chunks))
		out.append("""		<td align="center"><span class="%s">%s</span></td>""" % ("" if hsc.mchunks==0 else "MISSING", decimal_number_html(hsc.mchunks)))
		out.append("""		<td align="center"><span class="%s">%s</span></td>""" % ("" if hsc.ugchunks==0 else "UNDERGOAL", decimal_number_html(hsc.ugchunks)))
		out.append("""	</tr>""")
		if hsc.msgbuffleng>0:
			if hsc.msgbuffleng==FSTEST_INFO_STR_LIMIT:
				out.append("""	<tr><th colspan="10">Important messages (truncated, there is more...):</th></tr>""")
			else:
				out.append("""	<tr><th colspan="10">Important messages:</th></tr>""")
			out.append("""	<tr>""")
			out.append("""		<td colspan="10" align="left"><pre>%s</pre></td>""" % (htmlentities(hsc.datastr)))
			out.append("""	</tr>""")
	else:
		out.append("""	<tr>""")
		out.append("""		<td colspan="10" align="center">No data, self-check loop not finished yet</td>""" )
		out.append("""	</tr>""")
	out.append("""</table>""")

	if show_mfiles:
		import views.missingfiles
		out.append("\n".join(views.missingfiles.render(dp, fields, vld)))
		out.append(""" </td></tr>""")
		out.append("""</table>""")
	
	return out